"""CalendarTreeview class
Ttk treeview class with added functions.
"""
import tkinter as tk
from tkinter import ttk
from customtkinter import *
from bson.errors import InvalidId
from bson.objectid import ObjectId
from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.models.interval import Interval
from pollenisatorgui.core.models.tags import Tag
from pollenisatorgui.core.models.ip import Ip
from pollenisatorgui.core.models.port import Port
from pollenisatorgui.core.models.scope import Scope
from pollenisatorgui.core.models.checkinstance import CheckInstance
from pollenisatorgui.core.models.wave import Wave
from pollenisatorgui.core.models.command import Command
from pollenisatorgui.core.views.intervalview import IntervalView
from pollenisatorgui.core.views.ipview import IpView
from pollenisatorgui.core.views.multipleipview import MultipleIpView
from pollenisatorgui.core.views.multiplescopeview import MultipleScopeView
from pollenisatorgui.core.views.multiselectionview import MultiSelectionView
from pollenisatorgui.core.views.portview import PortView
from pollenisatorgui.core.views.scopeview import ScopeView
from pollenisatorgui.core.views.toolview import ToolView
from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView
from pollenisatorgui.core.views.waveview import WaveView
from pollenisatorgui.core.views.defectview import DefectView
from pollenisatorgui.core.views.commandview import CommandView
from pollenisatorgui.core.views.checkitemview import CheckItemView
from pollenisatorgui.core.controllers.wavecontroller import WaveController
from pollenisatorgui.core.controllers.checkinstancecontroller import CheckInstanceController
from pollenisatorgui.core.controllers.portcontroller import PortController
from pollenisatorgui.core.controllers.scopecontroller import ScopeController
from pollenisatorgui.core.controllers.toolcontroller import ToolController
from pollenisatorgui.core.controllers.defectcontroller import DefectController
from pollenisatorgui.core.controllers.ipcontroller import IpController
from pollenisatorgui.core.controllers.intervalcontroller import IntervalController
from pollenisatorgui.core.controllers.commandcontroller import CommandController
from pollenisatorgui.core.controllers.checkitemcontroller import CheckItemController
from pollenisatorgui.core.application.dialogs.ChildDialogProgress import ChildDialogProgress
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.application.dialogs.ChildDialogCustomCommand import ChildDialogCustomCommand
from pollenisatorgui.core.application.dialogs.ChildDialogExportSelection import ChildDialogExportSelection
from pollenisatorgui.core.application.treeviews.PollenisatorTreeview import PollenisatorTreeview
import pollenisatorgui.core.components.utils as utils
from PIL import Image
from pollenisatorgui.core.components.logger_config import logger


class PentestTreeview(PollenisatorTreeview):
    """Inherit PollenisatorTreeview.
    Ttk treeview class with added functions to handle the main view objects.
    """

    def __init__(self, appli, parentFrame):
        """
        Args:
            appli: a reference to the main Application object.
            parentFrame: the parent tkinter window object.
        """
        super().__init__(appli, parentFrame)
        self._hidden = []  # Hidden objects references
        self._hidden_lkp = {}  # Hidden objects ids, avoid time consuming list comprehension
        self.tagsMenu = None  # Menu bar for tag changing
        self.tagsVars = None  # list of tag values
        self.contextualMenu = None  #  contextual menu to open on right click
        self.waves_node = None  # the wave parent node on the treeview
        self.ips_node = None  # the IPs parent node on the treeview
        self.openedViewFrameId = None  # id of opened objects in the view frame.
        self.commands_node = None  # parent of all commands nodes
        self.lazylimit = 10000
        self.lazyload = False
        self.first_load = True
        DataManager.getInstance().attach(self)

    def initUI(self):
        """Initialize the user interface widgets and binds them."""
        self._initContextualsMenus()
        self.heading('#0', text='Object', anchor=tk.W)
        self.column('#0', stretch=tk.YES, minwidth=300, width=300)
        self.bind("<Button-3>", self.doPopup)
        self.bind("<Button-2>", self.doPopupTag)
        self.bind("<<TreeviewSelect>>", self.onTreeviewSelect)
        self.bind("<<TreeviewOpen>>", self.onTreeviewOpen)
        self.bind("<Double-Button-1>", self.onTreeviewDoubleClick)
        self.bind('<Delete>', self.deleteSelected)
        self.bind('h', self.hideSelection)
        self.bind("<Shift-Down>", self.openNextSameTypeNode)
        self.bind("<Shift-Up>", self.openPrevSameTypeNode)
        self.filterFrame = CTkFrame(self.parentFrame, height=0)
        self.filterFrameInterior = CTkFrame(self.filterFrame, height=0)
        self.initFiltersFrame(self.filterFrameInterior)
        self.filterFrame.grid(row=2, column=0, columnspan=2)

        self.filter_open_image = CTkImage(Image.open(utils.getIcon("up-arrow.png")))
        self.filter_button = CTkButton(self.parentFrame, text="", image=self.filter_open_image,command=self.filter_toogle)
        self.filter_open = True
        self.filter_toogle(self.filter_open)

    def reopen(self,event=None):
        self.onTreeviewSelect()

    def _place_button_filter(self):
        self.filter_button.place(in_= self, rely=1.01, relx=0.5, anchor=tk.S, bordermode="inside")

    def filter_toogle(self, status=None):
        if status is None:
            self.filter_open = not self.filter_open
        else:
            self.filter_open = status
        if self.filter_open:
            self.filter_image = CTkImage(Image.open(utils.getIcon("down-arrow.png")))
            self.filterFrame.configure(height=self.filterFrameInterior.winfo_reqheight())
            self.filterFrameInterior.pack(fill=tk.X, padx=5, pady =10)
        else:
            self.filterFrameInterior.pack_forget()
            self.filter_image = CTkImage(Image.open(utils.getIcon("up-arrow.png")))
            self.filterFrame.configure(height=0)
        self.filter_button.configure(image=self.filter_image)
        

    def initFiltersFrame(self, frame):
        """Populate the filter frame with cool widgets"""
        settings = Settings()
        settings.reloadLocalSettings()
        form = FormPanel( pady=0, padx=0, fill="y")
        show_only_todo = settings.is_show_only_todo()
        hide_oos = settings.is_hide_oos()
        self.check_hide_oos = form.addFormSwitch("OOS", "Hide Out Of Scope hosts", hide_oos, command=self.hideOOSSwap, pady=0, padx=0, side="top", anchor="w")
        self.check_show_only_todo = form.addFormSwitch("Todos", "Show only todos", show_only_todo, command=self.showTodoSwap, pady=0, padx=0, side="top", anchor="w")
        show_only_manual = settings.is_show_only_manual()
        self.check_show_only_manual = form.addFormSwitch("Manuals", "Show only manual checks", show_only_manual, command=self.showManualSwap, pady=0, padx=0, side="top", anchor="w")
        form.constructView(frame)
        
    def hideOOSSwap(self, event=None):
        val = self.check_hide_oos.getValue()
        settings = Settings()
        settings.local_settings["hide_oos"] = val
        settings.saveLocalSettings()
        if val == 0:
            self.unfilter("filter_oos")
        else:
            self.filter_oos()

    def checklistViewSwap(self, event=None):
        self.refresh()
        self.filter_empty_nodes()

    def showTodoSwap(self, event=None):
        val = self.check_show_only_todo.getValue() 
        settings = Settings()
        settings.local_settings["show_only_todo"] = val
        settings.saveLocalSettings()
        if val == 0:
            self.unfilter("filter_empty")
            self.unfilter("filter_todo")
        else:
            self.filter_todo()

    def filter_empty_nodes(self):
        for key, values in self.views.items():
            view = values["view"]
            if not isinstance(view, CheckItemView):
                continue
            iid = view.controller.getDbId()
            try:
                if len(self.get_children(str(iid))) == 0:
                    view.hide("filter_empty")
            except tk.TclError:
                pass

    def filter_todo(self):
        for values in self.views.values():
            view = values["view"]
            if not isinstance(view, CheckInstanceView):
                continue
            status = view.controller.getStatus()
            if status == "":
                status = view.controller.model.status
            if status == "":
                status = "todo"
            if status != "todo":
                view.hide("filter_todo")
        self.filter_empty_nodes()
    
    def filter_oos(self):
        for key, values in self.views.items():
            view = values["view"]
            if not isinstance(view, IpView):
                continue
            tags = self.item(key)["tags"]
            if "OOS" in tags:
                view.hide("filter_oos")
    
    def showManualSwap(self, event=None):
        val = self.check_show_only_manual.getValue() 
        settings = Settings()
        settings.local_settings["show_only_manual"] = val
        settings.saveLocalSettings()
        if val == 0:
            self.unfilter("filter_empty")
            self.unfilter("filter_manual")
        else:
            self.filter_manual()
        
    def filter_manual(self):
        for values in self.views.values():
            view = values["view"]
            if not isinstance(view, CheckInstanceView) and not isinstance(view, CheckItemView):
                continue
            if view.controller.isAuto():
                view.hide("filter_manual")
        self.filter_empty_nodes()
    
    def unfilter(self, reason):
        self.unhide(reason)


    def getRows(self, startNode=''):
        """Returns all child nodes of the given startNode iid as a list.
        Args:
            - startNode: node to recursively get children. Default to '' which is rhe treeview root.
        Returns:
            - List of all children iid of given node.
        """
        myRows = []
        children = self.get_children(startNode)
        for child in children:
            myRows.append(child)
            myRows += self.getRows(child)
        return myRows

    def openPrevSameTypeNode(self, _event=None):
        """Open the first node of the same type above the currently selected object on the tree view.
        Args:
            - _event: not used but mandatory
        Return:
            return the string "break" to stop processing the event
        """
        fromNode = self.selection()
        if fromNode:
            fromNode = fromNode[0]
            rows = self.getRows()
            pos = rows.index(fromNode)
            if pos-1 > 0:
                view_o = self.getViewFromId(fromNode)
                classToFind = view_o.__class__
                for row_nth in range(pos-1, -1, -1):
                    row = rows[row_nth]
                    cmp_view_o = self.getViewFromId(row)
                    if classToFind == cmp_view_o.__class__:
                        self.see(row)
                        self.focus(row)
                        self.selection_set(row)
                        return "break"
        return "break"

    def openNextSameTypeNode(self, _event=None):
        """Open the first node of the same type below the currently selected object on the tree view.
        Args:
            - _event: not used but mandatory
        Return:
            return the string "break" to stop processing the event
        """
        fromNode = self.selection()
        if fromNode:
            fromNode = fromNode[0]
            rows = self.getRows()
            pos = rows.index(fromNode)
            if pos+1 < len(rows):
                view_o = self.getViewFromId(fromNode)
                classToFind = view_o.__class__
                for row in rows[pos+1:]:
                    cmp_view_o = self.getViewFromId(row)
                    if classToFind == cmp_view_o.__class__:
                        self.see(row)
                        self.focus(row)
                        self.selection_set(row)
                        return "break"
        return "break"

    def doPopup(self, event):
        """Called when a right click is received by the tree view.
        Open the contextual menu at the clicked position.
        Args:
            - event: sent automatically though an event on treeview
        """
        if self.contextualMenu is not None:
           self.popupFocusOut()
        self.contextualMenu.selection = self.identify(
            "item", event.x, event.y)
        view = self.getViewFromId(str(self.contextualMenu.selection))
        additonal_commands = {}
        apiclient = APIClient.getInstance()
        datamanager = DataManager.getInstance()
        if view is None:
            item = self.contextualMenu.selection
            if str(item) == "waves":
                view = WaveView(self, self.appli.viewframe,
                                       self.appli, WaveController(Wave()))
            elif str(item) == "mycommands":
                user = apiclient.getUser()
                view = CommandView(
                            self, self.appli.viewframe, self.appli, CommandController(Command({"indb": apiclient.getCurrentPentest(), "owners": [user]})))
            elif str(item) == "commands":
                view = CommandView(
                        self, self.appli.viewframe, self.appli, CommandController(Command({"indb": apiclient.getCurrentPentest()})))
            elif str(item) == "ips":
                view = MultipleIpView(
                        self, self.appli.viewframe, self.appli, IpController(Ip()))
            elif "intervals" in str(item):
                wave = datamanager.get("waves",ObjectId(IntervalView.treeviewListIdToDb(item)))
                view = IntervalView(self, self.appli.viewframe, self.appli, IntervalController(
                        Interval().initialize(wave.wave)))
            elif "scopes" in str(item):
                wave = datamanager.get("waves",ObjectId(ScopeView.treeviewListIdToDb(item)))
                view = MultipleScopeView(
                            self, self.appli.viewframe, self.appli, ScopeController(Scope().initialize(wave.wave)))
        if view is not None:
            additonal_commands |= view.getAdditionalContextualCommands()
            
        if self.appli.searchMode:
            additonal_commands["Show in full tree"] = self.showInTreeview

        self._initContextualsMenus(additonal_commands)
        
        super().doPopup(event)

    def doPopupTag(self, event):
        """Called when a middle click is received by the tree view.
        Open the tag menu at the clicked position.
        Args:
            - event: sent automatically though an event on treeview
        """
        self.tagsMenu.selection = self.identify(
            "item", event.x, event.y)
        # display the popup menu
        try:
            self.tagsMenu.tk_popup(event.x_root, event.y_root)
        finally:
            # make sure to release the grab (Tk 8.0a1 only)
            self.tagsMenu.grab_release()
        self.tagsMenu.focus_set()
        self.tagsMenu.bind('<FocusOut>', self.popupFocusOutTag)

    def popupFocusOutTag(self, _=None):
        """Called when the tag contextual menu is unfocused.
        Close the tag contextual menu.
        """
        self.tagsMenu.unpost()

    def attackInTerminal(self, check_view):
        self.appli.open_terminal(str(check_view.controller.getDbId()), check_view.controller.target_repr)

    def insertViewById(self, node_id):
        view = self.views.get(node_id)
        if view is not None:
            view["view"].addInTreeview(insert_parents=True, addChildren=False)

    def showInTreeview(self, _=None):
        """Unfilter the treeview and focus the node stored in the contextualMenu.selection variable0
        Also select it.
        """
        node = str(self.contextualMenu.selection)
        self.appli.endSearch()
        try:
            self.delete("search")
        except tk.TclError:
            pass
        view = self.views.get(node)["view"]
        if view:
            self.unhideAll()
            view.addInTreeview(insert_parents=True)
            self.see(node)
            self.focus(node)
            self.selection_set(node)

    def tagClicked(self, name):
        """Callback for an event. If the function was called directly it would not work.
        Args:
            - name: the tag name clicked
        """
        return lambda: self.setTagFromMenubar(name)

    def _initContextualsMenus(self, additional_commands=None):
        """
        Create the contextual menu
        """
        additional_commands = {} if additional_commands is None else additional_commands
        self.contextualMenu = utils.craftMenuWithStyle(self.parentFrame)
        for command_name, command_func in additional_commands.items():
            self.contextualMenu.add_command(
                label=command_name, command=command_func)
        self.contextualMenu.add_command(
            label="Export selection", command=self.exportSelection)
        self.contextualMenu.add_separator()
        for module in self.appli.modules:
            if callable(getattr(module["object"], "_initContextualsMenus", None)):
                menu = module["object"]._initContextualsMenus(self.parentFrame)
                self.contextualMenu.add_cascade(
                    label=module["name"].strip(), menu=menu)
        self.tagsMenu = utils.craftMenuWithStyle(self.parentFrame)
        tags = Settings.getTags()
        listOfLambdas = [self.tagClicked(tag) for tag in list(tags.keys())]
        for i, val in enumerate(tags):
            self.tagsMenu.add_command(
                label=val, command=listOfLambdas[i])
        #self.contextualMenu.add_cascade(label="Tags", menu=self.tagsMenu)
        self.contextualMenu.add_command(
            label="Sort children", command=self.sort)
        self.contextualMenu.add_command(
            label="Expand", command=self.expand)
        self.contextualMenu.add_command(
            label="Collapse", command=self.collapse)
        self.contextualMenu.add_command(
            label="Hide", command=self.hideAndUpdate)
        self.contextualMenu.add_command(
            label="Unhide children", command=lambda: self.unhideNodeChildren("*"))
        self.contextualMenu.add_separator()
        self.contextualMenu.add_command(
            label="Debug", command=self.showDebug)
        self.contextualMenu.add_command(
            label="Close", command=self.closeMenu)
        return self.contextualMenu

    def setTagFromMenubar(self, name):
        """
        Change the tags of every selected object in the treeview to the one selected in the tag contextual menu.
        Args:
            - name: the tag name clicked
        """
        for selected in self.selection():
            view_o = self.getViewFromId(selected)
            view_o.controller.setTags([name])

    def closeMenu(self, _event=None):
        """
        Close the contextual menu. Does nothing, just an empty callback
        Args:
            - _event: not used but mandatory
        """
        # Do nothing and close
        return

    def modelToView(self, collection, model):
        view = None
        if not collection.endswith("s"):
            collection += "s"
        if collection == "tools":
            view = ToolView(self, self.appli.viewframe,
                            self.appli, ToolController(model))
        elif collection == "waves":
            view = WaveView(self, self.appli.viewframe,
                            self.appli, WaveController(model))
        elif collection == "scopes":
            view = ScopeView(self, self.appli.viewframe,
                                self.appli, ScopeController(model))
            
        elif collection == "ports":
            view = PortView(self, self.appli.viewframe,
                            self.appli, PortController(model))
        elif collection == "ips":
            view = IpView(self, self.appli.viewframe,
                            self.appli, IpController(model))
        elif collection == "intervals":
            view = IntervalView(self, self.appli.viewframe,
                                self.appli, IntervalController(model))
        elif collection == "defects":
            view = DefectView(self, self.appli.viewframe,
                                self.appli, DefectController(model))
        elif collection == "commands":
            view = CommandView(self, self.appli.viewframe,
                                self.appli, CommandController(model))
        elif collection == "checkinstances":
                check_i = model
                view = CheckInstanceView(self, self.appli.viewframe,
                                         self.appli, CheckInstanceController(check_i))
        elif collection == "checkitems":
                check_i = model
                view = CheckItemView(self, self.appli.viewframe,
                                         self.appli, CheckItemController(check_i))
                                        
                # parent_view = CheckItemView(self, self.appli.viewframe,
                #                          self.appli, CheckItemController(check_i.check_m))
                # parent_view.addInTreeview()
        return view

    def update_received(self, dataManager, notif, obj, old_obj):
        apiclient = APIClient.getInstance()
        if not apiclient.getCurrentPentest() != "":
            return
        if apiclient.getCurrentPentest() != notif["db"]:
            return
        collection = notif["collection"]
        # Delete
        if notif["action"] == "delete":
            try:
                self.delete(ObjectId(notif["iid"]))
            except tk.TclError:
                pass  # item was not inserted in the treeview
            if str(notif["iid"]) == self.openedViewFrameId:
                self.event_generate("<F6>") # event generation is handled in main thread (tkinter is not thread safe, so refrshing of view must be done through events)
        # Insert
        if collection == "tags" and notif["action"] == "insert":
            iid = obj.item_id
            try:
                self.item(str(iid), tags=obj.tags)
            except tk.TclError:
                pass
            return
        if notif["action"] == "insert":
            view = None
            if obj is None:
                return
            model = obj
            view = self.modelToView(collection, model)
            try:
                if view is not None:
                    view.addInTreeview()
                    view.insertReceived()
            except tk.TclError:
                pass
        elif notif["action"] == "insert_many":
            view = None
            if not obj:
                return
            for model in obj:
                view = self.modelToView(collection, model)
                try:
                    if view is not None:
                        view.addInTreeview(insert_parents=True)
                        view.insertReceived()
                except tk.TclError:
                    pass
        elif notif["action"] == "update_many":
            if not obj:
                return
            for model in obj:
                view = self.getViewFromId(model.getId())
                if view is not None:
                    tags = view.controller.getTags()
                    try:
                        self.item(str(model.getId()), tags=tags, image=view.getIcon())
                        if "hidden" in tags:
                            self.hide("tags")
                    except tk.TclError:
                        if view is not None:
                            view.addInTreeview(insert_parents=True)
                    if view is not None:
                        view.updateReceived(obj, old_obj)
                    if model.getId() == self.openedViewFrameId:
                        self.event_generate("<F6>")
        if notif["action"] == "update":
            try:
                if obj is None:
                    return
                self.updateView(str(notif["iid"]), obj)
                view = self.getViewFromId(str(notif["iid"]))
                if view is not None:
                    item = self.item(str(notif["iid"]))
                    tags = view.controller.getTags()
                    self.item(str(notif["iid"]), tags=tags, image=view.getIcon())
                    if "hidden" in tags:
                        self.hide("tags")
            except tk.TclError:
                if view is not None:
                    view.addInTreeview()
            
            if view is not None:
                view.updateReceived(obj, old_obj)
            if str(notif["iid"]) == self.openedViewFrameId:
                self.event_generate("<F6>")

    def exportSelection(self, _event=None):
        """
        Popup a window to help a user to export some data from treeview.
        Args:
            _event: not used but mandatory
        """
        values = set()
        for selected in self.selection():
            view_o = self.getViewFromId(selected)
            if view_o is not None:
                data = view_o.controller.getData()
                for key in data.keys():
                    if key == "infos":
                        for info_keys in data["infos"]:
                            values.add("infos."+info_keys)
                    else:
                        values.add(key)
        dialog = ChildDialogExportSelection(self, values)
        self.wait_window(dialog.app)
        if isinstance(dialog.rvalue, list):
            fields_to_export = dialog.rvalue
            f = tk.filedialog.asksaveasfilename(parent=self, defaultextension=".csv")
            # asksaveasfile return `None` if dialog closed with "cancel".
            if f is None or len(f) == 0:
                return
            csv_filename = str(f)
            with open(csv_filename, mode='w') as f:
                f.write(", ".join(fields_to_export)+"\n")
                for selected in self.selection():
                    view_o = self.getViewFromId(selected)
                    if view_o is not None:
                        line = []
                        modelData = view_o.controller.getData()
                        for field in fields_to_export:
                            if field.startswith("infos."):
                                line.append(
                                    str(modelData.get("infos", {}).get(field[6:], "")))
                            else:
                                line.append(str(modelData.get(field, "")))
                        f.write(", ".join(line)+"\n")
            dialog = ChildDialogQuestion(
                self, "Export completed", "Your export just finished : "+csv_filename+".\n Do you want to open this folder ?")
            self.wait_window(dialog.app)
            if dialog.rvalue == "Yes":
                utils.openPathForUser(csv_filename, folder_only=True)

    def onTreeviewOpen(self, event=None):
        """Called when an item is opened, lazy load its children.
        """
        if not self.lazyload:
            return
        opened = event.widget.focus()
        if opened == "" or opened is None:
            return
        objView = self.getViewFromId(str(opened))
        if objView is not None:
            objView.opened()
        return 
    
    def onTreeviewSelect(self, event=None):
        """Called when a line is selected on the treeview
        Open the selected object view on the view frame.
        Args:
            _event: not used but mandatory
        """
        selection = self.selection()
        if len(selection) == 1:
            item = super().onTreeviewSelect(event)
            if isinstance(item, str):
                if item.startswith("multi|"):
                    objView = self.getViewFromId(str(item))
                    self.openModifyWindowOf(item)
                #self._openInsertWindow(item)
            else:
                objView = self.getViewFromId(str(item))
                objView.opened()
                self.item(selection, open=True)
                self.openModifyWindowOf(item)
        elif len(selection) > 1:
            prev_type = None
            all_the_same = True
            for item in selection:
                items = {}
                try:
                    item_iid = ObjectId(item)
                    objView = self.getViewFromId(str(item_iid))
                    if objView is not None:
                        items[str(item_iid)] = objView
                        curr_type = objView.__class__
                        if prev_type is not None and prev_type != curr_type:
                            all_the_same = False
                            break
                        prev_type = curr_type
                except InvalidId:
                    break
            if all_the_same and prev_type.multiview_class is not None:
                multiView = prev_type.multiview_class(
                    self, self.appli.viewframe, self.appli)
            else:
            # Multi select:
                multiView = MultiSelectionView(
                    self, self.appli.viewframe, self.appli)
            for widget in self.appli.viewframe.winfo_children():
                widget.destroy()
            multiView.form.clear()
            multiView.openModifyWindow()

    def onTreeviewDoubleClick(self,  event=None):
        selection = self.selection()
        if len(selection) == 1:
            item = super().onTreeviewSelect(event)
            objView = self.getViewFromId(str(item))
            if objView is not None:
                if hasattr(objView, "onDoubleClick") and callable(objView.onDoubleClick):
                    objView.onDoubleClick()

    def _openInsertWindow(self, item):
        apiclient = APIClient.getInstance()
        self.saveState(apiclient.getCurrentPentest())
        if str(item) == "waves":
            objView = WaveView(self, self.appli.viewframe,
                                       self.appli, WaveController(Wave()))
            objView.openInsertWindow()
        elif str(item) == "mycommands":
            user = apiclient.getUser()
            objView = CommandView(
                        self, self.appli.viewframe, self.appli, CommandController(Command({"indb": apiclient.getCurrentPentest(), "owners": [user]})))
            objView.openInsertWindow()
        elif str(item) == "commands":
            objView = CommandView(
                        self, self.appli.viewframe, self.appli, CommandController(Command({"indb": apiclient.getCurrentPentest()})))
            objView.openInsertWindow()
        elif str(item) == "ips":
            objView = MultipleIpView(
                        self, self.appli.viewframe, self.appli, IpController(Ip()))
            objView.openInsertWindow()
        elif "intervals" in str(item):
            wave = Wave.fetchObject(
                        {"_id": ObjectId(IntervalView.treeviewListIdToDb(item))})
            objView = IntervalView(self, self.appli.viewframe, self.appli, IntervalController(
                        Interval().initialize(wave.wave)))
            objView.openInsertWindow()
        elif "scopes" in str(item):
            wave = Wave.fetchObject(
                        {"_id": ObjectId(ScopeView.treeviewListIdToDb(item))})
            objView = MultipleScopeView(
                        self, self.appli.viewframe, self.appli, ScopeController(Scope().initialize(wave.wave)))
            objView.openInsertWindow()

    def load(self, searchModel=None):
        """
        Load the treeview with database information
        Args:
            searchModel: (DEPRECATED) a search object default to None
        """
        for widget in self.appli.viewframe.winfo_children():
            widget.destroy()
        # Reattach hidden as get_children won't get them otherwise
        self.unhideAll()
        # Clear the tree
        apiclient = APIClient.getInstance()
        self.heading("#0", text=apiclient.getCurrentPentestName())
        self.delete(*self.get_children())
        
        self._hidden = []
        self._hidden_lkp = {}
        if searchModel is None or searchModel.query == "":
            self._load()
        else:
            viewsFound = searchModel.getViews(
                self, self.appli.viewframe, self.appli)
            for viewFound in viewsFound:
                viewFound.addInTreeview('', addChildren=False)
        self.configureTags()
        self.loadState(apiclient.getCurrentPentest())

    # def changeSize(self, event):
    #     self._place_button_filter()


    def refresh(self, force=True):
        """Alias to load function"""
        self.after(0, self._place_button_filter)
        if force or self.first_load:
            self.first_load = False
            self.unfilterAll()
            #DataManager.getInstance().load(collections=["ip", "port","wave","scope","interval", "command"],forceReload=force)
            
            #DataManager.getInstance().load(collections=["ip", "wave","scope","interval", "command"],forceReload=force)
            self.load()
        
    def _load(self):
        """ 
        Load the treeview with database information
        """
        self.appli.settings.reloadSettings()

        checklistview = self.appli.settings.is_checklist_view()

        apiclient = APIClient.getInstance()
        nbObjects = apiclient.count("checkinstances")
        if nbObjects > self.lazylimit and not self.lazyload:
            self.lazyload = True
            tk.messagebox.showwarning(
                "Warning Many Objects detected", "Big pentest detected. Lazy loading will be used. (required data will be downloaded on request)")
        dialog = ChildDialogProgress(self.appli, "Loading "+str(
            apiclient.getCurrentPentestName()), "Opening "+str(apiclient.getCurrentPentestName()) + ". Please wait for a few seconds.", speed=10, progress_mode="determinate")
        self.unhideAll()
        for child in self.get_children():
            self.delete(child)   
        self._hidden = []
        self._hidden_lkp = {}
        dataManager = DataManager.getInstance()
        self.appli.statusbar.reset()
        dialog.update(msg="Insertings cheatsheets")
        # self.commands_node = self.insert(
        #     self.commands_node, "end", "mycommands", text="Commands", image=CommandView.getClassIcon())        # Loading every category separatly is faster than recursivly.
        # This is due to cursor.next function calls in pymongo
        # Adding wave objects
        # LOAD BIG PENTEST OBJECT
        dialog.update(msg="Insertings ips")
        # Adding ip objects
        try:
            self.ips_node = self.insert("", "end", str(
                "ips"), text="IPs", image=IpView.getClassIcon())
        except tk.TclError as e:
            print(e)
        dialog.update(msg="Insertings checkinstances node")
        try:
            self.commands_node = self.insert(
            "", "end", "commands", text="Commands", image=CommandView.getClassIcon())
        except tk.TclError as e:
            pass
        
        if checklistview:
            checks = list(dataManager.get("checkitem", '*'))
            apiclient = APIClient.getInstance()
            checks.sort(key=lambda x: x.priority)
            for check in checks:
                count = apiclient.count("checkinstances", {"check_iid": str(check._id)})
                if count > 0:
                    check_vw = CheckItemView(
                        self, self.appli.viewframe, self.appli, CheckItemController(check))
                    check_vw.addInTreeview(None, with_category=True, addChildren=False, count_children=count)
            if not self.lazyload:
                tags = dataManager.find("tags", {"item_type": "checkinstances"}, fetch_on_none=True)
                tag_lkp = {}
                for tag in tags:
                    tag_lkp[str(tag.item_id)] = tag
                checkinstances = dataManager.get("checkinstance", '*')
                check_lkp = {str(checkinstance.getId()): checkinstance for checkinstance in checkinstances}
                targets_repr = apiclient.getCheckInstanceRepr(list(check_lkp.keys()))
                for checkinstance in checkinstances:
                    check_o = CheckInstanceController(checkinstance)
                    check_o.target_repr = targets_repr.get(str(checkinstance.target_iid), "<Removed>")
                    checkinstance_vw = CheckInstanceView(
                        self, self.appli.viewframe, self.appli, check_o)
                    checkinstance_vw.addInTreeview(None, addChildren=False, tags=tag_lkp.get(str(checkinstance.getId()), []))
        
        else:
            try:
                self.waves_node = self.insert("", "end" , str(
                    "waves"), text="Scope options", image=WaveView.getClassIcon())
            
            except tk.TclError as e:
                pass
            
            
            
                #dialog.update()
            dialog.update(msg="Insertings commands")
            for command in list(dataManager.get("command", '*')):
                command_vw = CommandView(
                    self, self.appli.viewframe, self.appli, CommandController(command))
                command_vw.addInTreeview()
                #dialog.update()
            dialog.update(msg="Insertings waves ")
            waves = list(dataManager.get("wave", '*'))
            for wave in waves:
                wave_o = WaveController(wave)
                wave_vw = WaveView(
                    self, self.appli.viewframe, self.appli, wave_o)
                wave_vw.addInTreeview(self.waves_node, addChildren=False)
                
            # dialog.update()
            intervals = list(dataManager.get("interval", '*'))
            for interval in intervals:
                interval_o = IntervalController(interval)
                interval_vw = IntervalView(
                    self, self.appli.viewframe, self.appli, interval_o)
                interval_vw.addInTreeview(None, addChildren=False)
                #dialog.update()
            scopes = list(dataManager.get("scope", '*'))
            for scope in scopes:
                scope_o = ScopeController(scope)
                scope_vw = ScopeView(
                    self, self.appli.viewframe, self.appli, scope_o)
                scope_vw.addInTreeview(None, addChildren=False)
            dialog.update(msg="Insertings ips")
            
            if not checklistview:
                ips = list(dataManager.get("ip", '*'))
                for ip in ips:
                    ip_o = IpController(ip)
                    ip_vw = IpView(self, self.appli.viewframe, self.appli, ip_o)
                    ip_vw.addInTreeview(self.ips_node, addChildren=False)
                
            if not self.lazyload:
                dialog.update(msg="Insertings ports")
                    #dialog.update()
                # Adding port objects
                ports = list(dataManager.get("port", '*'))
                for port in ports:
                    port_o = PortController(port)
                    port_vw = PortView(
                        self, self.appli.viewframe, self.appli, port_o)
                    port_vw.addInTreeview(None, addChildren=False)
        # LOAD MODULE PENTEST OBJECTS        
        for module in self.appli.modules:
            if hasattr(module["object"], "onTreeviewLoad"):
                module["object"].onTreeviewLoad(checklistview, self.lazyload)
        # ADD DEFECT, CHECKINSTANCE OBJECTS and TOOL IF NOT LAZYLOAD AND NOT CHECKLISTVIEW
        if not checklistview:
            if not self.lazyload:
                #dialog.update()
                # Adding defect objects
                dialog.update(msg="Insertings defects")
                defects = list(dataManager.get("defect", '*'))
                for defect in defects:
                    defect_o = DefectController(defect)
                    defect_vw = DefectView(
                        self, self.appli.viewframe, self.appli, defect_o)
                    defect_vw.addInTreeview(None)
                dialog.update(msg="Insertings check instances")
                apiclient = APIClient.getInstance()

                checks = list(dataManager.get("checkinstance", '*'))
                check_lkp = {check.getId(): check for check in checks}
                targets_repr = apiclient.getCheckInstanceRepr(list(check_lkp.keys()))
                for check in checks:
                    check_o = CheckInstanceController(check)
                    check_o.target_repr = targets_repr.get(str(check.target_iid), "<Removed>")
                    check_vw = CheckInstanceView(
                        self, self.appli.viewframe, self.appli, check_o)
                    check_vw.addInTreeview(None, refresh_status=False)
                dialog.update(msg="Insertings tools")
                tools = list(dataManager.get("tools",'*'))
                for tool in tools:
                    tool_o = ToolController(tool)
                    tool_vw = ToolView(
                        self, self.appli.viewframe, self.appli, tool_o)
                    tool_vw.addInTreeview(None)
        dialog.update(msg="Preparing treeview...")
        if not checklistview:
            # Adding Tools objects
            self.open("scopes")
            if not self.lazyload:
                self.open("ips")
            self.sort(self.ips_node)
        else:
            try:
                self.hide(self.ips_node, "checklist_view")
                self.hide(self.commands_node, "checklist_view")
                self.hide(self.waves_node, "checklist_view")
            except tk.TclError as e:
                pass
        dialog.destroy()


    def openModifyWindowOf(self, dbId):
        """
        Retrieve the View of the database id given and open the modifying form for its model.
        Args:
            dbId: the Mongo Id to open the modification form on.
        """
        # self.resetTags()
        objView = self.getViewFromId(str(dbId))
        if objView is not None:
            for widget in self.appli.viewframe.winfo_children():
                widget.destroy()
            objView.form.clear()
            self.openedViewFrameId = str(dbId)
            objView.openModifyWindow(treevw="pentest")
           

    # def unhide(self, node=None, reason=""):
    #     """Unhide children of given node.
    #     Args:
    #         - node: the node which we want to unhide the children.
    #                 If this value is None, use the contextualMenu.selection value
    #                 Default to None.
    #     """
    #     nodeToUnhideChildren = str(
    #         self.contextualMenu.selection) if node is None else node

    #     for hidden in self._hidden[::-1]:
    #         itemId = hidden[0]
    #         parentId = '' if hidden[1] is None else hidden[1]
    #         if str(parentId) == str(nodeToUnhideChildren):
    #             view_o = self.getViewFromId(str(itemId))
    #             if view_o is not None:
    #                 view_o.controller.delTag("hidden")
    #             try:
    #                 self.reattach(itemId, parentId, 0)
    #             except tk.TclError:
    #                 pass  # Le noeud a ete supprime entre temps

    def hideAndUpdate(self):
        """Hide object with contextualMenu attached in the treeview and store this effect in its tags."""
        self.hide(None, "manual_hide", True)

    def showDebug(self):
        """Hide object with contextualMenu attached in the treeview and store this effect in its tags."""

        node = str(self.contextualMenu.selection)
        view_o = self.getViewFromId(node)
        msg = str(view_o)+"\n"+str(view_o.controller.getDbId())+"\n"
        data = view_o.controller.getData()
        msg += str(data)
        dialog = ChildDialogAskText(self, "Debug", default=msg)
        self.wait_window(dialog.app)

    def hideSelection(self, _event=None):
        """ Hide selected objects in the treeview and store this effect in their tags."""
        selectedNodes = self.selection()
        for node in selectedNodes:
            self.hide(node, "manual_hide", True)

    def hide(self, node=None, reason="manual_hide", updateTags=False):
        """Hide given node object in the treeview and can store this effect in its tag.
        Args:
            - node: node to hide. If none is given, the contextualMenu.selection value will be used.
                    Default to None.
            - updateTags: mark the object as hidden in its tags. Default to False
        """
        if node is None and not hasattr(self.contextualMenu, "selection"):
            return
        nodeToHide = str(
            self.contextualMenu.selection) if node is None else node
        view_o = self.getViewFromId(nodeToHide)
        if view_o is not None:
            parent = view_o.getParentNode()
        else:
            parent = self.parent(node)
        hidd = self._hidden_lkp.get(str(nodeToHide), None)
        if hidd is not None:
            self._hidden[hidd[0]][2].append(reason)
        else:
            self._hidden.append([nodeToHide, parent, [reason]])
            self._hidden_lkp[str(nodeToHide)] = (len(self._hidden)-1, parent)
            
            
        
        
        try:
            if updateTags and view_o is not None:
                view_o.controller.addTag("hidden")
            self.detach(nodeToHide)
        except tk.TclError:
            pass

    def parent(self, itemiid):
        data = self._hidden_lkp.get(str(itemiid))
        if data is None:
            return super().parent(itemiid)
        parentId = '' if data[1] is None else data[1]
        return parentId
        # for hidden in self._hidden[::-1]:
        #     itemId = hidden[0]
        #     parentId = '' if hidden[1] is None else hidden[1]
        #     if str(itemId) == str(itemiid):
        #         return parentId
        #return super().parent(itemiid)

    def showItem(self, item):
        self.see(str(item))
        self.selection_set(str(item))
        self.focus(str(item))



from bson import ObjectId
from pollenisatorgui.core.application.dialogs.ChildDialogAskFile import ChildDialogAskFile
from pollenisatorgui.core.components.apiclient import APIClient
import pollenisatorgui.core.components.utils as utils
import os


def main(apiclient, appli, **kwargs):
    cme_path = utils.which_expand_alias("cme")
    if not cme_path:
        return False, "binary 'cme' is not in the PATH."
    APIClient.setInstance(apiclient)
    ip = ""
    if kwargs.get("target_type").lower() == "computer":
        computer_info =  apiclient.find("computers", {"type":"computer", "_id":ObjectId(kwargs.get("target_iid", ""))}, False)
        if computer_info is not None:
            ip = computer_info.get("ip", "")
    if ip == "" or ip is None:
        return False, "No ip given"
    dialog = ChildDialogAskFile(None, "List of users to test:")
    
    if dialog.rvalue is None:
        return False, "No user list given"
    file_name = dialog.rvalue
    if not os.path.exists(file_name):
        return False, "Userlist given not found."
    
    iid = appli.launch_in_terminal(kwargs.get("default_target",None), "CME kerberos users", f"{cme_path} ldap {ip} -u {file_name} -p '' -k")
    return True, f"Launched cme in terminal, if an Invalid principal syntax is raised, you did not setup the krb5.conf or DNS for your env"



"""Widget with an entry to type a date and a calendar button to choose a date."""

import tkinter as tk
from tkinter import ttk
import datetime as datetime
from pollenisatorgui.core.forms.form import Form
from PIL import ImageTk, Image
from customtkinter import *
from pollenisatorgui.core.components.utils import getIcon
from pollenisatorgui.core.application.dialogs.ChildDialogDate import ChildDialogDate
from pollenisatorgui.core.application.pollenisatorentry import PopoEntry

class FormDate(Form):
    """
    Form field representing a date.
    Default setted values: 
        state="readonly"
        if pack : padx = pady = 5, side = "right"
        if grid: row = column = 0 sticky = "west"
    """
    def __init__(self, name, root, default="", dateformat='%d/%m/%Y %H:%M:%S', **kwargs):
        """
        Constructor for a form checkbox

        Args:
            name: the date name (id).
            root: the tkinter root window
            default: a list of string that should be prechecked if in the choice list.
            dateformat: a date format as a string see datetime.strptime documentation.
            kwargs: same keyword args as you would give to CTkFrame
        """
        super().__init__(name)
        self.dateformat = dateformat
        self.default = default
        self.root = root
        self.kwargs = kwargs
        self.entry = None
        FormDate.img_class = CTkImage(
            Image.open(getIcon("date.png")))

    def constructView(self, parent):
        """
        Create the date view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        self.parent = parent
        self.val = tk.StringVar()
        frame = CTkFrame(parent.panel)
        self.entry = PopoEntry(frame, textvariable=self.val)
        self.val.set(self.default)
        
        self.entry.grid(row=0, column=0)
        datepicker = CTkLabel(frame, text="", image=FormDate.img_class)
        datepicker.grid(row=0,column=1, padx=5)
        datepicker.bind("<Button-1>", self.showDatePicker)
        for bind, bind_call in self.getKw("binds", {}).items():
            if bind == "<<OnDateModified>>":
                self.val.trace("w", bind_call)
            else:
                self.entry.bind(bind, bind_call)
        if parent.gridLayout:
            frame.grid(row=self.getKw("row", 0), column=self.getKw("column", 0), sticky=self.getKw("sticky", tk.W), **self.kwargs)
        else:
            frame.pack(side=self.getKw("side", "right"), padx=self.getKw("padx", 10), pady=self.getKw("pady", 5), **self.kwargs)

    def showDatePicker(self, _event=None):
        """Callback to start displaying the date picker calendar window
        Args:
            _event: mandatory but not used
        """
        dialog = ChildDialogDate(self.root)
        self.root.wait_window(dialog.app)
        if dialog.rvalue is not None:
            datestr = datetime.datetime.strftime(dialog.rvalue, self.dateformat)
            self.val.set(datestr)

   
    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the date as string text.
        """
        return self.val.get()

    def checkForm(self):
        """
        Check if this form is correctly filled. Check with the dateformat given in constructorn or "None".

        Returns:
            {
                "correct": True if the form is correctly filled, False otherwise.
                "msg": A message indicating what is not correctly filled.
            }
        """
        value = self.getValue()
        if value == "None":
            return True, ""
        try:
            # exception is raised if string cannot be converted with dateformat
            datetime.datetime.strptime(value, self.dateformat)
        except ValueError:
            return False, self.name+" does not respect the date format ("+str(self.dateformat)+")"
        return True, ""

    def setFocus(self):
        """Set the focus to the ttk entry part of the widget.
        """
        self.entry.focus_set()


"""Hold functions to interact with the settings"""
import json
import os
import tkinter as tk
import tkinter.messagebox
import tkinter.ttk as ttk
from shutil import which

from customtkinter import *
from PIL import Image

import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform
from pollenisatorgui.core.components.apiclient import APIClient, ErrorHTTP
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.application.pollenisatorentry import PopoEntry

class Settings:
    """
    Represents the settings of pollenisator.
    There are three level of settings:
        * local settings: stored in a file under ../../config/settings.cfg
        * pentest db settings: stored in the pentest database under settings collection
        * global settings: stored in the pollenisator database under settings collection
    """
    tags_cache = None
    __pentest_types = None
    def __init__(self):
        """
        Load the tree types of settings and stores them in dictionnaries
        """
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.confdir = os.path.join(dir_path, "../../config/settings.cfg")

        self.local_settings = {}
        try:
            with open(self.confdir, mode="r") as f:
                self.local_settings = json.loads(f.read())
        except json.JSONDecodeError:
            self.local_settings = {}
        except IOError:
            self.local_settings = {}
        self.db_settings = {}
        self.global_settings = {}
        self.text_pentest_types = None
        self.text_tags = None
        self.text_db_tags = None
        self.box_pentest_type = None
        self.visual_include_domains_with_ip_in_scope = None
        self.visual_include_domains_with_topdomain_in_scope = None
        self.visual_search_show_hidden = None
        self.visual_search_exact_match = None
        self.visual_include_all_domains = None
        self.visual_dark_mode = None
        #self.text_pentesters = None
        self.box_favorite_term = None
        self.visual_trap_commands = None
        self.pentesters_treevw = None
        from pollenisatorgui.core.components.datamanager import DataManager
        DataManager.getInstance().attach(self)

    @classmethod
    def getPentestTags(cls):
        apiclient = APIClient.getInstance()
        db_tags = apiclient.find("settings", {"key":"tags"}, False)
        if db_tags is None:
            db_tags = {}
        else:
            db_tags = db_tags["value"]
        return db_tags

    @classmethod
    def getTags(cls, onlyGlobal=False, **kwargs):
        """
        Returns tags defined in settings.
        Returns:
            If none are defined returns default
            otherwise returns a dict with defined key values
        """
        apiclient = APIClient.getInstance()
        if kwargs.get("ignoreCache", False): #Check if ignore cache is true
            cls.tags_cache = None
        if cls.tags_cache is not None and not onlyGlobal:
            return cls.tags_cache
        cls.tags_cache = {"todo":{"color":"orange", "level":"todo"}, "pwned":{"color":"red", "level":"high"}, "Interesting":{"color":"dark green", "level":"medium"}, "Uninteresting":{"color":"sky blue", "level":"low"}, "neutral":{"color":"transparent", "level":""}}
        try:
            global_tags = apiclient.getSettings({"key": "tags"})
        except ErrorHTTP:
            global_tags = None
        if global_tags is not None:
            if isinstance(global_tags["value"], dict):
               global_tags = global_tags["value"]
            elif isinstance(global_tags["value"], str):
                global_tags = json.loads(global_tags["value"])
        if global_tags is None:
            global_tags = {}
        if isinstance(global_tags, str):
            global_tags = json.loads(global_tags)
        if not onlyGlobal:
            try:
                db_tags = cls.getPentestTags()
            except ErrorHTTP:
                db_tags = {}
            cls.tags_cache = {**global_tags, **db_tags}
            return cls.tags_cache
        return global_tags

    @classmethod
    def getPentestTypes(cls):
        """
        Returns pentest types and associeted defect type defined in settings.
        Returns:
            If none are defined returns {"Web":["Base", "Application", "Data", "Policy"], "LAN":["Infrastructure", "Active Directory", "Data", "Policy"]}
            otherwise returns a dict with defined key values
        """
        apiclient = APIClient.getInstance()
        if cls.__pentest_types is None:
            pentest_types = apiclient.getSettings({"key": "pentest_types"})
            if pentest_types is not None:
                if isinstance(pentest_types["value"], str):
                    cls.__pentest_types = json.loads(pentest_types["value"])
                elif isinstance(pentest_types["value"], dict):
                    cls.__pentest_types = pentest_types["value"]
                else:
                    cls.__pentest_types = {"Web":["Base", "Application", "Data", "Policy"], "LAN":["Infrastructure", "Active Directory", "Data", "Policy"]}
            else:
                cls.__pentest_types = {"Web":["Base", "Application", "Data", "Policy"], "LAN":["Infrastructure", "Active Directory", "Data", "Policy"]}
        return cls.__pentest_types

    def getClientName(self):
        return self.db_settings.get("client_name", "")
    
    def getMissionName(self):
        return self.db_settings.get("mission_name", "")

   
    
    def is_hide_oos(self):
        return self.local_settings.get("hide_oos", False)
    
    def is_checklist_view(self):
        return self.local_settings.get("checklist_view", True)

    def is_show_only_todo(self):
        return self.local_settings.get("show_only_todo", False)

    def is_show_only_manual(self):
        return self.local_settings.get("show_only_manual", False)

    def isTrapCommand(self):
        return self.local_settings.get("trap_commands", True)

    def is_dark_mode(self):
        return self.local_settings.get("dark_mode", False)

    def setTrapCommand(self):
        self.local_settings["trap_commands"] = self.visual_trap_commands.get()
        self.saveLocalSettings()


    
    def reloadLocalSettings(self):
        """
        Reload local settings from local conf file
        """
        try:
            with open(self.confdir, mode="r") as f:
                self.local_settings = json.loads(f.read())
        except json.JSONDecodeError:
            self.local_settings = {}
        except IOError:
            self.local_settings = {}

    def _reloadDbSettings(self):
        """
        Reload pentest database settings from pentest database
        """
        apiclient = APIClient.getInstance()
        dbSettings = apiclient.find("settings", {})
        if dbSettings is None:
            dbSettings = {}
        self.__class__.tags_cache = None
        for settings_dict in dbSettings:
            try:
                self.db_settings[settings_dict["key"]] = settings_dict["value"]
            except KeyError:
                pass

    def _reloadGlobalSettings(self):
        """
        Reload pentest database settings from pollenisator database
        """
        apiclient = APIClient.getInstance()
        globalSettings = apiclient.getSettings()
        self.__class__.tags_cache = None
        for settings_dict in globalSettings:
            self.global_settings[settings_dict["key"]] = settings_dict["value"]

    def reloadSettings(self):
        """
        Reload local, database and global settings.
        """
        self.reloadLocalSettings()
        self._reloadDbSettings()
        self._reloadGlobalSettings()

    def reloadUI(self):
        """
        Reload all settings and refresh view with values
        """
        self.reloadSettings()
        self.visual_include_all_domains.set(
            self.db_settings.get("include_all_domains", False))
        self.visual_include_domains_with_ip_in_scope.set(
            self.db_settings.get("include_domains_with_ip_in_scope", False))
        self.visual_include_domains_with_topdomain_in_scope.set(
            self.db_settings.get("include_domains_with_topdomain_in_scope", False))
        self.text_pentest_name.delete("0", tk.END)
        apiclient = APIClient.getInstance()
        self.text_pentest_name.insert("0", apiclient.getCurrentPentestName())
        self.text_mission_name.delete("0", tk.END)
        self.text_mission_name.insert("0",
            self.db_settings.get("mission_name", ""))
        self.text_client_name.delete("0", tk.END)
        self.text_client_name.insert("0",
            self.db_settings.get("client_name", ""))
        self.combo_lang.set(self.db_settings.get("lang", "en"))
        self.visual_search_show_hidden.set(
            self.local_settings.get("search_show_hidden", True))
        self.visual_search_exact_match.set(
            self.local_settings.get("search_exact_match", False))
        self.visual_dark_mode.set(
            self.local_settings.get("dark_mode", False))
        self.visual_trap_commands.set(self.local_settings.get("trap_commands", False))
        #self.text_pentesters.delete('1.0', tk.END)
        #self.text_pentesters.insert(
        #    tk.INSERT, "\n".join(
        #        self.db_settings.get("pentesters", [])))
        self.pentesters_treevw.reset()
        pentesters_as_list = self.db_settings.get("pentesters", [""])
        dict_for_tw = {}
        for pentester in pentesters_as_list:
            dict_for_tw[pentester] = tuple([pentester])
        self.pentesters_treevw.recurse_insert(dict_for_tw)
        self.box_pentest_type.set(self.db_settings.get("pentest_type", "None"))
        self.text_pentest_types.delete('1.0', tk.END)
        pentestTypes = Settings.getPentestTypes()
        buffer = ""
        for pentestType, pentestTypeDefectTypes in pentestTypes.items():
            buffer += pentestType +" : "+ (", ".join(pentestTypeDefectTypes))+"\n"
        self.text_pentest_types.insert(
            tk.INSERT, buffer)
        self.text_tags.delete('1.0', tk.END)
        tagsRegistered = Settings.getTags(onlyGlobal=True)
        buffer = json.dumps(tagsRegistered, indent=4)
        self.text_tags.insert(
            tk.INSERT, buffer)
        self.text_db_tags.delete('1.0', tk.END)
        tagsPentestRegistered = Settings.getPentestTags()
        buffer = json.dumps(tagsPentestRegistered, indent=4)
        self.text_db_tags.insert(
            tk.INSERT, buffer)
        
    def saveLocalSettings(self):
        """
        Save local settings to conf file
        """
        with open(self.confdir, mode="w") as f:
            f.write(json.dumps(self.local_settings))

    def setPentestSetting(self, key, val):
        apiclient = APIClient.getInstance()
        self.db_settings[key] = val
        apiclient.updatePentestSetting({"key":key, "value": json.dumps(val)})

    def savePentestSettings(self):
        apiclient = APIClient.getInstance()
        settings = apiclient.find("settings")
        existing_settings = {}
        for setting in settings:
            existing_settings[setting["key"]] = setting
        for k, v in self.db_settings.items():
            if k in existing_settings:
                if k == "tags":
                    for line_key, line_value in v.items():
                        tag, tag_infos = line_key, line_value
                        if tag not in existing_settings["tags"]["value"]:
                            apiclient.registerTag(apiclient.getCurrentPentest(), tag, tag_infos["color"], tag_infos["level"])
                        else:
                            apiclient.updateTag(tag, tag_infos["color"], tag_infos["level"], False)
                    for tag in existing_settings["tags"].get("value",{}):
                        if tag not in v:
                            apiclient.unregisterTag(apiclient.getCurrentPentest(), tag)
                else:
                    apiclient.updatePentestSetting({"key":k, "value": json.dumps(v)})

    def save(self):
        """
        Save all the settings (local, database and global)
        """
        apiclient = APIClient.getInstance()

        for k, v in self.global_settings.items():
            if apiclient.getSettings({"key": k}) is None:
                apiclient.createSetting(k, v)
            else:
                apiclient.updateSetting(k, v)
        self.savePentestSettings()
        self.saveLocalSettings()
        self.reloadUI()

    def initUI(self, parent):
        """Create settings widgets and initialize them
        Args:
            parent: parent tkinter container widget"""
        if self.visual_include_all_domains is not None:  # Already built
            self.reloadUI()
            return
        self.parent = parent
        self.image_save = CTkImage(Image.open(utils.getIcon("save.png")))

        self.settingsFrame = ScrollableFrameXPlateform(parent)
        parent.columnconfigure(0, weight=1)
        parent.rowconfigure(0, weight=1)

        self.visual_include_all_domains = tk.BooleanVar()
        self.visual_dark_mode = tk.BooleanVar()
        self.visual_include_domains_with_ip_in_scope = tk.BooleanVar()
        self.visual_include_domains_with_topdomain_in_scope = tk.BooleanVar()
        self.visual_search_show_hidden = tk.BooleanVar()
        self.visual_search_exact_match = tk.BooleanVar()
        self.visual_trap_commands = tk.BooleanVar()
        chkbox_dark_mode = CTkSwitch(self.settingsFrame, text="Dark mode",
                                                     variable=self.visual_dark_mode)
        chkbox_dark_mode.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.W)
        lbl_domains = ttk.LabelFrame(
            self.settingsFrame, text="Discovered domains options")
        
        chkbox_include_domains_with_ip_in_scope = CTkSwitch(lbl_domains, text="Check if discovered subdomains ips are in scope",
                                                                  variable=self.visual_include_domains_with_ip_in_scope)
        chkbox_include_domains_with_ip_in_scope.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.W)

        chkbox_include_domains_with_topdomain_in_scope = CTkSwitch(lbl_domains, text="Check if discovered subdomains have a top domain already in scope",
                                                                         variable=self.visual_include_domains_with_topdomain_in_scope)
        chkbox_include_domains_with_topdomain_in_scope.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.W)
        chkbox_include_all_domains = CTkSwitch(lbl_domains, text="/!\\ Include every domain found in scope",
                                                     variable=self.visual_include_all_domains)
        chkbox_include_all_domains.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.W)
        lbl_domains.pack(padx=10, pady=10, side=tk.TOP,
                         anchor=tk.CENTER, fill=tk.X, expand=tk.YES)
        frame_term = ttk.LabelFrame(self.settingsFrame, text="Local terminal configuration")
        chkbox_trap_commands = CTkSwitch(frame_term, text="Trap every command (instead of using pollex)", variable=self.visual_trap_commands)
        chkbox_trap_commands.pack(side=tk.TOP, pady=5)
        frame_term.pack(padx=10, pady=10, side=tk.TOP,
                           anchor=tk.CENTER, fill=tk.X, expand=tk.YES)
        lbl_SearchBar = ttk.LabelFrame(self.settingsFrame, text="Search settings")
        
        chkbox_search_show_hidden = CTkSwitch(lbl_SearchBar, text="Show hidden objects",
                                                    variable=self.visual_search_show_hidden)
        chkbox_search_show_hidden.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.W)
        chkbox_search_exact_match = CTkSwitch(lbl_SearchBar, text="Exact match",
                                                    variable=self.visual_search_exact_match)
        chkbox_search_exact_match.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.W)
        lbl_SearchBar.pack(padx=10, pady=10, side=tk.TOP,
                           anchor=tk.CENTER, fill=tk.X, expand=tk.YES)
        lblframe_pentest_params = ttk.LabelFrame(
            self.settingsFrame, text="Pentest parameters")
        lbl_pentest_name = CTkLabel(lblframe_pentest_params, text="Pentest name")
        lbl_pentest_name.grid(row=0, column=0, sticky=tk.E)
        self.text_pentest_name = PopoEntry(lblframe_pentest_params)
        self.text_pentest_name.grid(row=0, column=1, sticky=tk.W)
        lbl_client_name = CTkLabel(lblframe_pentest_params, text="Client's name")
        lbl_client_name.grid(row=1, column=0, sticky=tk.E)
        self.text_client_name = PopoEntry(lblframe_pentest_params)
        self.text_client_name.grid(row=1, column=1, sticky=tk.W)
        
        lbl_mision_name = CTkLabel(lblframe_pentest_params, text="Mission name")
        lbl_mision_name.grid(row=2, column=0, sticky=tk.E)
        self.text_mission_name = PopoEntry(lblframe_pentest_params)
        self.text_mission_name.grid(row=2, column=1, sticky=tk.W)
        lbl_report_lang =  CTkLabel(lblframe_pentest_params, text="Report language")
        lbl_report_lang.grid(row=3, column=0, sticky=tk.E)
        langs = APIClient.getInstance().getLangList()
        self.combo_lang = CTkComboBox(lblframe_pentest_params, values=langs)
        self.combo_lang.grid(row=3, column=1, sticky=tk.W)
        lbl_pentest_type = CTkLabel(
            lblframe_pentest_params, text="Pentest type")
        lbl_pentest_type.grid(row=4, column=0, sticky=tk.E)
        self.box_pentest_type = CTkComboBox(
            lblframe_pentest_params, values=tuple(Settings.getPentestTypes().keys()))
        self.box_pentest_type.grid(row=4, column=1, sticky=tk.W)
        # self.text_pentesters = CTkTextbox(
        #     lblframe_pentest_params, height=3, font = ("Sans", 10))
        # lbl_pentesters = CTkLabel(
        #     lblframe_pentest_params, text="Pentester names")
        # lbl_pentesters.grid(row=2, column=0, sticky=tk.E)
        # self.text_pentesters.grid(row=2, column=1, sticky=tk.W, pady=5)
        lblframe_pentest_params.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.CENTER, fill=tk.X, expand=tk.YES)
        form_pentesters_panel = CTkFrame(self.settingsFrame)
        self.form_pentesters = FormPanel(side=tk.TOP, fill=tk.X, pady=5)
        self.form_pentesters.addFormSearchBar("Pentester search", self.searchCallback, self.form_pentesters, side=tk.TOP)
        self.pentesters_treevw = self.form_pentesters.addFormTreevw(
            "Additional pentesters names", ["Additional pentesters names"], (""), add_empty_row=True, height=20, width=100, pady=5, side=tk.LEFT)
        
        self.form_pentesters.constructView(form_pentesters_panel)
        form_pentesters_panel.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.CENTER, fill=tk.X, expand=tk.YES)
        lblframe_global_params = ttk.LabelFrame(
            self.settingsFrame, text="Other parameters")
        lblframe_global_params.columnconfigure(1, weight=2)
        lbl_pentest_types = CTkLabel(
            lblframe_global_params, text="Pentests possible types")
        lbl_pentest_types.grid(row=0, column=0, sticky=tk.E, pady=10, padx=5)
        self.text_pentest_types = CTkTextbox(
            lblframe_global_params,  wrap="word")
        self.text_pentest_types.grid(row=0, column=1, sticky=tk.W+tk.E, pady=10, padx=5)
        lbl_tags = CTkLabel(
            lblframe_global_params, text="Registered tags")
        lbl_tags.grid(row=1, column=0, sticky=tk.E, pady=10, padx=5)
        self.text_tags = CTkTextbox(
            lblframe_global_params,wrap="word")
        self.text_tags.grid(row=1, column=1, sticky=tk.W+tk.E)
        lbl_db_tags = CTkLabel(
            lblframe_global_params, text="Pentest only tags")
        lbl_db_tags.grid(row=2, column=0, sticky=tk.E, pady=10, padx=5)
        self.text_db_tags = CTkTextbox(
            lblframe_global_params, wrap="word")
        
        self.text_db_tags.grid(row=2, column=1, sticky=tk.W+tk.E, pady=10)
        lblframe_global_params.pack(
            padx=10, pady=10, side=tk.TOP, anchor=tk.CENTER, fill=tk.X, expand=tk.YES)
        btn_save = CTkButton(parent, text="Save", command=self.on_ok, image=self.image_save)
        btn_save.grid(row=3, column=0, padx=10, pady=10, sticky="s")
        self.settingsFrame.grid(column=0, row=0, sticky="nsew")
        #self.settingsFrame.pack(fill=tk.BOTH, expand=1)
        #self.reloadUI()

    def searchCallback(self, searchreq):
        apiclient = APIClient.getInstance()
        users = apiclient.searchUsers(searchreq)
        if users is None:
            return [], "Invalid response from API"
        ret = [{"TITLE":user, "additional pentesters names":{"text":user}} for user in users]
        return ret, ""

    def on_ok(self):
        """Callback on click save button. loads some data and calls save.
        Args:
            parent: parent tkinter container widget"""
        info = "Settings saved"
        self.db_settings["include_all_domains"] = self.visual_include_all_domains.get(
        ) == 1
        self.db_settings["include_domains_with_ip_in_scope"] = self.visual_include_domains_with_ip_in_scope.get(
        ) == 1
        pentest_name = self.text_pentest_name.get().strip()
        apiclient = APIClient.getInstance()
        if apiclient.getCurrentPentestName() != pentest_name:
            apiclient.updatePentest(pentest_name)
        self.db_settings["mission_name"] = self.text_mission_name.get().strip()
        self.db_settings["client_name"] = self.text_client_name.get().strip()
        self.db_settings["lang"] = self.combo_lang.get().strip()
        self.db_settings["pentest_type"] = self.box_pentest_type.get()
        self.db_settings["include_domains_with_topdomain_in_scope"] = self.visual_include_domains_with_topdomain_in_scope.get(
        ) == 1
        self.db_settings["pentesters"] = []
        form_values = self.form_pentesters.getValue()
        form_values_as_dicts = ViewElement.list_tuple_to_dict(form_values)
        self.db_settings["pentesters"] = [x for x in form_values_as_dicts["Additional pentesters names"] if x.strip() != ""]
        self.db_settings["tags"] = json.loads(self.text_db_tags.get('1.0', tk.END))
       

        self.local_settings["search_show_hidden"] = self.visual_search_show_hidden.get(
        ) == 1
        self.local_settings["search_exact_match"] = self.visual_search_exact_match.get(
        ) == 1
        old = self.local_settings.get("dark_mode", False)
        self.local_settings["dark_mode"] = self.visual_dark_mode.get(
        ) == 1
        if old != self.local_settings["dark_mode"]:
            info = "Dark mode changed. You need to restart the application to see the changes."
        self.local_settings["trap_commands"] = self.visual_trap_commands.get() == 1
        self.global_settings["pentest_types"] = {}
        for type_of_pentest in self.text_pentest_types.get('1.0', tk.END).split(
                "\n"):
            if type_of_pentest.strip() != "":
                line_splitted = type_of_pentest.strip().split(":")
                if len(line_splitted) == 2:
                    typesOfDefects = list(map(lambda x: x.strip(), line_splitted[1].split(",")))
                    self.global_settings["pentest_types"][line_splitted[0].strip()] = typesOfDefects
        self.global_settings["tags"] = json.loads(self.text_tags.get('1.0', tk.END))
        self.save()
        tkinter.messagebox.showinfo(
            "Settings", info)

    def getPentestType(self):
        """Return selected database pentest type.
        Returns:
            Open database pentest type. string "None" if not defined"""
        return self.db_settings.get("pentest_type", "None")

    def getPentesters(self):
        """Return a list of pentesters registered for open pentest database
        Returns:
            List of pentesters names"""
        return self.db_settings.get("pentesters", [])

    def update_received(self, dataManager, notif, obj, old_obj):
        if notif["collection"] != "settings":
            return
        if notif["db"] == "pollenisator":
            self._reloadGlobalSettings()
        else:
            self._reloadDbSettings()

    
        

        


# """Dashboard module to display pentest info"""
# import tkinter as tk
# import tkinter.ttk as ttk
# from customtkinter import *

# from PIL import Image, ImageTk

# from pollenisatorgui.core.application.dialogs.ChildDialogProgress import ChildDialogProgress
# from pollenisatorgui.core.components.apiclient import APIClient
# from pollenisatorgui.core.models.ip import Ip
# from pollenisatorgui.core.models.port import Port
# from pollenisatorgui.core.models.tool import Tool
# import pollenisatorgui.core.components.utils as utils

# from pollenisatorgui.modules.module import Module

# class DashBoard(Module):
#     """
#     Shows information about ongoing pentest. 
#     """
#     iconName = "tab_dashboard.png"
#     tabName = " Dashboard "
#     def __init__(self, parent, settings):
#         """
#         Constructor
#         """
#         super().__init__()
#         self.dashboardFrame = None
#         self.parent = None
#         self.treevw = None
#         self.style = None
#         self.ips = None
#         self.ports = None
#         self.tools = None

#         iconPath = utils.getIconDir()
#         self.icons = {}
#         self.icons["tool"] = CTkImage(
#             Image.open(iconPath+"tool.png"))
#         self.icons["cross"] = CTkImage(
#             Image.open(iconPath+"cross.png"))
#         self.icons["running"] = CTkImage(
#             Image.open(iconPath+"running.png"))
#         self.icons["done"] = CTkImage(
#             Image.open(iconPath+"done_tool.png"))
#         self.icons["error"] = CTkImage(
#             Image.open(iconPath+"error_tool.png"))
#         self.icons["ready"] = CTkImage(
#             Image.open(iconPath+"waiting.png"))
#         self.icons["Not ready"] = CTkImage(
#             Image.open(iconPath+"cross.png"))
    
#     def open(self):
#         apiclient = APIClient.getInstance()
#         if apiclient.getCurrentPentest() is not None:
#             self.refreshUI()
#         return True

#     def refreshUI(self):
#         """
#         Reload data and display them
#         """
#         self.loadData()
#         self.displayData()

#     def loadData(self):
#         """
#         Fetch data from database
#         """
#         self.ports = Port.fetchObjects({})
#         self.ips = Ip.fetchObjects({})
#         self.tools = Tool.fetchObjects({})

#     def displayData(self):
#         """
#         Display loaded data in treeviews
#         """
#         dialog = ChildDialogProgress(self.parent, "Loading dashboard ",
#                                      "Refreshing dashboard. Please wait for a few seconds.", 200, "determinate")
#         dialog.show(10)

#         # Reset Ip treeview
#         for children in self.treevw.get_children():
#             self.treevw.delete(children)
#         dialog.update(1)
#         listOfip = []
#         for ip in self.ips:
#             servicesCount = len([x for x in Port.fetchObjects({"ip": ip.ip})])
#             listOfip.append((ip.ip, servicesCount))
#         dialog.update(2)
#         listOfip.sort(key=lambda tup: tup[1], reverse=True)
#         for i in range(len(listOfip)):
#             self.treevw.insert(
#                 '', 'end', i, text=listOfip[i][0], values=(listOfip[i][1]))
#         dialog.update(3)
#         # Reset Port treeview
#         for children in self.treevwport.get_children():
#             self.treevwport.delete(children)
#         dialog.update(4)
#         portCounts = {}
#         for port in self.ports:
#             if port.port not in portCounts.keys():
#                 portCounts[port.port] = 1
#             else:
#                 portCounts[port.port] += 1
#         dialog.update(5)

#         port_id = 0
#         # Ordering dictionnary
#         portCounts = {k: v for k, v in sorted(
#             portCounts.items(), key=lambda item: item[1], reverse=True)}
#         for portCount in portCounts:
#             self.treevwport.insert('', 'end', port_id, text=str(
#                 portCount), values=(portCounts[portCount]))
#             port_id += 1
#         dialog.update(6)
#         # Tool part
#         # Reset Tools treeview
#         for children in self.treevwtools.get_children():
#             self.treevwtools.delete(children)
#         dialog.update(7)
#         listOfTools = [_ for _ in self.tools]
#         listOfTools.sort(key=lambda x: x.status, reverse=True)

#         result = APIClient.getInstance().aggregate("tools",
#                                                        [
#                                                            {
#                                                                "$group":
#                                                                {
#                                                                    "_id": {"name": "$name", "status": "$status", "wave": "$wave"}, "count": {"$sum": 1}
#                                                                }
#                                                            }
#                                                        ]
#                                                        )

#         result = [_ for _ in result]
#         tools_dashboard = {}
#         for tool_result in result:
#             tool_id = tool_result["_id"].get("wave","")+"::"+tool_result["_id"]["name"]
#             tools_dashboard[tool_id] = tools_dashboard.get(tool_id, {})
#             tool_status = list(tool_result["_id"].get("status", "ready"))
#             if not tool_status:
#                 tool_status = ["ready"]
#             tools_dashboard[tool_id][tool_status[0]] = tool_result["count"]
#         dialog.update(8)
#         for tool_id in sorted(list(tools_dashboard.keys())):
#             self.treevwtools.insert('', 'end', None, text=str(
#                 tool_id), values=(tools_dashboard[tool_id].get("ready", 0), tools_dashboard[tool_id].get("running", 0), tools_dashboard[tool_id].get("done", 0), tools_dashboard[tool_id].get("error", 0)+tools_dashboard[tool_id].get("timedout", 0)))
#         dialog.update(9)
#         # Defect Part
#         # reset defect TW
#         for children in self.treevwDefaults.get_children():
#             self.treevwDefaults.delete(children)

#         result = APIClient.getInstance().aggregate("defects",
#                                                        [
#                                                            {
#                                                                "$group":
#                                                                {
#                                                                    "_id": {"risk": "$risk", "type": "$type"}, "count": {"$sum": 1}
#                                                                }
#                                                            }
#                                                        ]
#                                                        )
#         dialog.update(10)
#         result = [_ for _ in result]
#         result.sort(key=lambda x: x["count"], reverse=True)
#         for defect in result:
#             defectRisk = defect["_id"]["risk"]
#             defectType = " ".join(defect["_id"]["type"])
#             defectCount = defect["count"]
#             self.treevwDefaults.insert('', 'end', None, text=str(
#                 defectRisk), values=(defectType, defectCount))
#         dialog.destroy()

#     def initUI(self, parent, nbk, treevw, tkApp):
#         """
#         Initialize Dashboard widgets
#         Args:
#             parent: its parent widget
#         """
#         if self.parent is not None:  # Already initialized
#             return
#         self.parent = parent
#         self.dashboardFrame = CTkFrame(parent)

#         self.rowHeight = 20
#         self.style = ttk.Style()
#         self.style.configure('DashBoard.Treeview', rowheight=self.rowHeight)

#         frameTwHosts = CTkFrame(self.dashboardFrame)
#         self.treevw = ttk.Treeview(
#             frameTwHosts, style='DashBoard.Treeview', height=5)
#         self.treevw['columns'] = ('services')
#         self.treevw.heading("#0", text='Hostname', anchor=tk.W)
#         self.treevw.column("#0", anchor=tk.W, width=150)
#         self.treevw.heading('services', text='Services')
#         self.treevw.column('services', anchor='center', width=40)
#         self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
#         scbVSel = CTkScrollbar(frameTwHosts,
#                                 orientation=tk.VERTICAL,
#                                 command=self.treevw.yview)
#         self.treevw.configure(yscrollcommand=scbVSel.set)
#         scbVSel.grid(row=0, column=1, sticky=tk.NS)
#         #frameTwHosts.pack(side=tk.TOP, fill=tk.X, padx=100, pady=10)
#         frameTwHosts.columnconfigure(0, weight=1)
#         frameTwHosts.rowconfigure(0, weight=1)
#         frameTwHosts.grid(row=0, column=0, sticky=tk.NSEW, padx=5, pady=10)

#         frameTwPports = CTkFrame(self.dashboardFrame)
#         self.treevwport = ttk.Treeview(
#             frameTwPports, style='DashBoard.Treeview', height=5)
#         self.treevwport['columns'] = ('amount')
#         self.treevwport.heading("#0", text='Port', anchor=tk.W)
#         self.treevwport.column("#0", anchor=tk.W, width=70)
#         self.treevwport.heading('amount', text='Amount')
#         self.treevwport.column('amount', anchor='center', width=40)
#         self.treevwport.grid(row=0, column=0, sticky=tk.NSEW)
#         scbVSel = CTkScrollbar(frameTwPports,
#                                 orientation=tk.VERTICAL,
#                                 command=self.treevwport.yview)
#         self.treevwport.configure(yscrollcommand=scbVSel.set)
#         scbVSel.grid(row=0, column=1, sticky=tk.NS)
#         frameTwPports.columnconfigure(0, weight=1)
#         frameTwPports.rowconfigure(0, weight=1)
#         frameTwPports.grid(row=0, column=1, sticky=tk.NSEW, padx=5, pady=10)

#         frameTools = CTkFrame(self.dashboardFrame)
#         self.treevwtools = ttk.Treeview(
#             frameTools, style='DashBoard.Treeview', height=10)
#         self.treevwtools['columns'] = ('ready', 'running', 'done', "error")
#         self.treevwtools.heading("#0", text='Tool', anchor=tk.W)
#         self.treevwtools.column("#0", anchor=tk.W, width=150)
#         self.treevwtools.heading('#1', text='Ready')
#         self.treevwtools.column('#1', anchor='center', width=10)
#         self.treevwtools.heading('#2', text='Running')
#         self.treevwtools.column('#2', anchor='center', width=10)
#         self.treevwtools.heading('#3', text='Done')
#         self.treevwtools.column('#3', anchor='center', width=10)
#         self.treevwtools.heading('#4', text='Error')
#         self.treevwtools.column('#4', anchor='center', width=10)
#         self.treevwtools.grid(row=0, column=0, sticky=tk.NSEW)
#         scbVSel = CTkScrollbar(frameTools,
#                                 orientation=tk.VERTICAL,
#                                 command=self.treevwtools.yview)
#         self.treevwtools.configure(yscrollcommand=scbVSel.set)
#         self.treevwtools.bind("<Motion>", self.mycallback)
#         scbVSel.grid(row=0, column=1, sticky=tk.NS)
#         frameTools.columnconfigure(0, weight=1)
#         frameTools.rowconfigure(0, weight=1)
#         frameTools.grid(row=1, sticky=tk.NSEW, padx=5, pady=5, columnspan=2)

#         frameDefaults = CTkFrame(self.dashboardFrame)
#         self.treevwDefaults = ttk.Treeview(
#             frameDefaults, style='DashBoard.Treeview', height=15)
#         self.treevwDefaults["columns"] = ('type', 'amount')
#         self.treevwDefaults.heading("#0", text='Risk', anchor=tk.W)
#         self.treevwDefaults.column("#0", anchor=tk.W, width=50)
#         self.treevwDefaults.heading('type', text='Type')
#         self.treevwDefaults.column('type', anchor='center', width=40)
#         self.treevwDefaults.heading('amount', text='Amount')
#         self.treevwDefaults.column('amount', anchor='center', width=40)
#         self.treevwDefaults.grid(row=0, column=0, sticky=tk.NSEW)
#         scbVSel = CTkScrollbar(frameDefaults,
#                                 orientation=tk.VERTICAL,
#                                 command=self.treevwDefaults.yview)
#         self.treevwDefaults.configure(yscrollcommand=scbVSel.set)
#         scbVSel.grid(row=0, column=1, sticky=tk.NS)
#         frameDefaults.columnconfigure(0, weight=1)
#         frameDefaults.rowconfigure(0, weight=1)
#         frameDefaults.grid(row=2, sticky=tk.NSEW, padx=5, columnspan=2, pady=5)
              
#         self.dashboardFrame.columnconfigure(0, weight=1)
#         self.dashboardFrame.columnconfigure(1, weight=1)
#         self.dashboardFrame.pack(padx=10, pady=10, fill=tk.BOTH, expand=True)
#         # self.refreshUI()

#     def mycallback(self, event):
#         _iid = self.treevwtools.identify_row(event.y)
#         self.treevwtools.item(_iid)



"""Provide useful functions"""
import multiprocessing
from pathlib import Path
import select
import sys
import os
import socket
import subprocess
import time
from datetime import datetime
from threading import Timer
import json
from netaddr import IPNetwork
import pty
from netaddr.core import AddrFormatError
from bson import ObjectId
import signal
from shutil import which
import shlex
import psutil
import tkinter  as tk
import tkinter.ttk as ttk
from customtkinter import *
from PIL import Image, ImageTk
from ttkwidgets import tooltips
from pollenisatorgui.core.components.logger_config import logger
import customtkinter
from os.path import expanduser
import screeninfo
settings = None



class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return "ObjectId|"+str(o)
        elif isinstance(o, datetime):
            return str(o)
        return json.JSONEncoder.default(self, o)

class JSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
        
    def object_hook(self, dct):
        for k,v in dct.items():
            if isinstance(v, list):
                new_lst = []
                for item in v:
                    if 'ObjectId|' in str(item):
                        new_lst.append(ObjectId(item.split('ObjectId|')[1]))
                    else:
                        new_lst.append(item)
                    dct[k] = new_lst
            else:
                if 'ObjectId|' in str(v):
                    dct[k] = ObjectId(v.split('ObjectId|')[1])
        return dct

def cacheSettings():
    """
    Cache the settings in a file.
    """
    global settings
    if settings is None:
        from pollenisatorgui.core.components.settings import Settings
        settings = Settings()
    return settings

def loadPlugin(pluginName):
    """
    Load a the plugin python corresponding to the given command name.
    The plugin must start with the command name and be located in plugins folder.
    Args:
        pluginName: the command name to load a plugin for

    Returns:
        return the module plugin loaded or default plugin if not found.
    """
    from pollenisatorgui.core.plugins.plugin import REGISTRY
    # Load plugins
    dir_path = os.path.dirname(os.path.realpath(__file__))
    path = os.path.join(dir_path, "../plugins/")
    # Load plugins
    sys.path.insert(0, path)
    try:
        # Dynamic import, raises ValueError if not found
        if not pluginName.endswith(".py"):
            pluginName += ".py"
        # trigger exception if plugin does not exist
        __import__(pluginName[:-3])
        return REGISTRY[pluginName[:-3]]  # removes the .py
    except ValueError:
        __import__("Default")
        return REGISTRY["Default"]
    except FileNotFoundError:
        __import__("Default")
        return REGISTRY["Default"]
    except ModuleNotFoundError:
        __import__("Default")
        return REGISTRY["Default"]

def getBackgroundColor():
    """
    Return the background color of the current style.
    Returns:
        the background color.
    """
    settings = cacheSettings()
    dark_mode = settings.is_dark_mode()
    colors = get_color_scheme(dark_mode)
    return colors.get("background")

def getBackgroundSecondColor():
    settings = cacheSettings()
    dark_mode = settings.is_dark_mode()
    colors = get_color_scheme(dark_mode)
    return colors.get("background_bis")

def getTextColor():
    """
    Return the background color of the current style.
    Returns:
        the background color.
    """
    settings = cacheSettings()
    dark_mode = settings.is_dark_mode()
    colors = get_color_scheme(dark_mode)
    return colors.get("text")

def getStrongColor():
    """
    Return the background color of the current style.
    Returns:
        the background color.
    """
    settings = cacheSettings()
    dark_mode = settings.is_dark_mode()
    colors = get_color_scheme(dark_mode)
    return colors.get("strong")

def getStrongActiveColor():
    """
    Return the background color of the current style.
    Returns:
        the background color.
    """
    settings = cacheSettings()
    dark_mode = settings.is_dark_mode()
    colors = get_color_scheme(dark_mode)
    return colors.get("strong_active")

def get_color_scheme(dark_mode=False):
    main_color = ("#5ac433", "#2FA572")[dark_mode]
    hover_color = ("#48a825", "#2FC572")[dark_mode]
    background_color = ("gray97", "gray17")[dark_mode]
    text_color = ("gray20", "#DCE4EE")[dark_mode]
    second_color = ("gray89", "dim gray")[dark_mode]
    strong_color = ('#3b75a8','#113759')[dark_mode]
    strong_active_color = ('#061b4e','#2768a1')[dark_mode]
    return {"main":main_color, "hover":hover_color, "background":background_color, "text":text_color, "background_bis":second_color,
            "strong":strong_color, "strong_active":strong_active_color}

def craftMenuWithStyle(menu_parent):
    """
    Craft a menu with the current style.
    Args:
        menu_parent: the parent menu to craft.

    Returns:
        the crafted menu.
    """
   
    menu = tk.Menu(menu_parent, tearoff=0, background= getStrongColor(), foreground="white", activebackground= getStrongActiveColor(), activeforeground="white")
    return menu

def setStyle(tkApp, dark_mode=False, _event=None):
    """
    Set the tk app style window widget style using ttk.Style
    Args:
        _event: not used but mandatory
    """
    if dark_mode:
        customtkinter.set_appearance_mode("dark")
    else:
        customtkinter.set_appearance_mode("light")

    style = ttk.Style(tkApp)
    style.theme_use("clam")
    
    colors = get_color_scheme(dark_mode)
    main_color = colors.get("main")
    hover_color = colors.get("hover") 
    background_color = colors.get("background")
    text_color = colors.get("text")
    background_bis_color = colors.get("background_bis")
    style.configure("Treeview.Heading", background=main_color,
                foreground="gray97", borderwidth=0)
    style.map('Treeview.Heading', background=[('active', hover_color)])
    style.configure("Treeview", background=background_color, fieldbackground=background_color, foreground=text_color)
    #style.configure("Treeview.Item")

    style.configure("TLabelframe", background=background_color,foreground=background_color,
                    labeloutside=False, bordercolor=main_color,relief=tk.SUNKEN)
    style.configure('TLabelframe.Label', background=main_color,
                    foreground=background_color, font=('Sans', '10', 'bold'))
    style.configure("TProgressbar",
                background=hover_color, foreground=hover_color, troughcolor=background_color, darkcolor=hover_color, lightcolor=hover_color)
    style.configure("Important.TFrame", background=main_color)
    style.configure("TFrame", background=background_color)
    style.configure("Debug.TFrame", background="red")
    style.configure("Important.TLabel", background=main_color, foreground=background_color)
    style.configure("Pagination.TLabel", background=background_color, foreground=main_color, font=('Sans', '10', 'underline') )
    style.configure("CurrentPagination.TLabel", background=background_color, foreground=main_color, font=('Sans', '10', 'bold') )
    style.map("Pagination.TLabel", foreground=[("active", hover_color)])
    style.map("CurrentPagination.TLabel", foreground=[("active", hover_color)] )
    style.configure("TLabel", background=background_color, foreground=text_color)
    style.configure("TCombobox", background=background_color)
    
    style.configure("TCheckbutton", background=background_color)

    style.configure("TButton", background=main_color,
                foreground=text_color, font=('Sans', '10', 'bold'), borderwidth=1)
    style.configure("icon.TButton", background=background_color, borderwidth=0)
    style.configure("iconbis.TButton", background=background_bis_color, borderwidth=0)
    style.configure("icon_white.TButton", background=main_color, borderwidth=0)

    style.configure("link.TButton", background=background_color, foreground=main_color, font=('Sans', '10', 'underline'), borderwidth=0)
    style.map('link.TButton', foreground=[('active', hover_color)], background=[('active', background_color)])
    style.configure("Notebook.TButton", background=main_color,
                foreground=text_color, font=('Sans', '10', 'bold'), borderwidth=0)
    style.configure("Notebook.TFrame", background=main_color)
    style.map('TButton', background=[('active', hover_color)])
    #  FIX tkinter tag_configure not showing colors   https://bugs.python.org/issue36468
    style.map('Treeview', foreground=fixedMap('foreground', style),
                background=fixedMap('background', style))
    tooltips.update_defaults({"background": getBackgroundColor()})

def fixedMap(option, style):
    """
    Fix color tag in treeview not appearing under some linux distros
    Args:
        option: the string option you want to affect on treeview ("background" for example)
        strle: the style object of ttk
    """
    # Fix for setting text colour for Tkinter 8.6.9
    # From: https://core.tcl.tk/tk/info/509cafafae
    #  FIX tkinter tag_configure not showing colors   https://bugs.python.org/issue36468
    # Returns the style map for 'option' with any styles starting with
    # ('!disabled', '!selected', ...) filtered out.

    # style.map() returns an empty list for missing options, so this
    # should be future-safe.
    return [elm for elm in style.map('Treeview', query_opt=option) if
            elm[:2] != ('!disabled', '!selected')]
                  
def isIp(domain_or_networks):
    """
    Check if the given scope string is a network ip or a domain.
    Args:
        domain_or_networks: the domain string or the network ipv4 range string
    Returns:
        Returns True if it is a network ipv4 range, False if it is a domain (any other possible case).
    """
    import re
    regex_network_ip = r"((?:[0-9]{1,3}\.){3}[0-9]{1,3})$"
    ipSearch = re.match(regex_network_ip, domain_or_networks)
    return ipSearch is not None


def isNetworkIp(domain_or_networks):
    """
    Check if the given scope string is a network ip or a domain.
    Args:
        domain_or_networks: the domain string or the network ipv4 range string
    Returns:
        Returns True if it is a network ipv4 range, False if it is a domain (any other possible case).
    """
    try:
        IPNetwork(domain_or_networks)
    except AddrFormatError:
        return False
    return True


def splitRange(rangeIp):
    """
    Check if the given range string is bigger than a /24, if it is, splits it in many /24.
    Args:
        rangeIp: network ipv4 range string
    Returns:
        Returns a list of IpNetwork objects corresponding to the range given as /24s.
        If the entry range is smaller than a /24 (like /25 ... /32) the list will be empty.
    """
    ip = IPNetwork(rangeIp)
    subnets = list(ip.subnet(24))
    return subnets


def resetUnfinishedTools():
    """
    Reset all tools running to a ready state. This is useful if a command was running on a worker and the auto scanning was interrupted.
    """
    # test all the cases if datef is defined or not.
    # Normally, only the first one is necessary
    from pollenisatorgui.core.models.tool import Tool
    tools = Tool.fetchObjects({"datef": "None", "scanner_ip": {"$ne": "None"}})
    for tool in tools:
        tool.markAsNotDone()
    tools = Tool.fetchObjects({"datef": "None", "dated": {"$ne": "None"}})
    for tool in tools:
        tool.markAsNotDone()
    tools = Tool.fetchObjects(
        {"datef": {"$exists": False}, "dated": {"$ne": "None"}})
    for tool in tools:
        tool.markAsNotDone()
    tools = Tool.fetchObjects(
        {"datef": {"$exists": False}, "scanner_ip": {"$ne": "None"}})
    for tool in tools:
        tool.markAsNotDone()


def stringToDate(datestring):
    """Converts a string with format '%d/%m/%Y %H:%M:%S' to a python date object.
    Args:
        datestring: Returns the date python object if the given string is successfully converted, None otherwise"""
    ret = None
    if isinstance(datestring, str):
        if datestring != "None":
            ret = datetime.strptime(
                datestring, '%d/%m/%Y %H:%M:%S')
    return ret

def dateToString(date):
    """Converts a date to a string with format '%d/%m/%Y %H:%M:%S'
    Args:
        datestring: Returns the str if the given datetime is successfully converted, None otherwise"""
    ret = None
    if isinstance(date, datetime):
        ret = date.strftime('%d/%m/%Y %H:%M:%S')
    return ret

def fitNowTime(dated, datef):
    """Check the current time on the machine is between the given start and end date.
    Args:
        dated: the starting date for the interval
        datef: the ending date for the interval
    Returns:
        True if the current time is between the given interval. False otherwise.
        If one of the args is None, returns False."""
    today = datetime.now()
    date_start = stringToDate(dated)
    date_end = stringToDate(datef)
    if date_start is None or date_end is None:
        return False
    return today > date_start and date_end > today

def handleProcKill(proc):
    logger.debug(f"Utils execute: handleProcKill {proc}")
    os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
    proc._killed = True

def check_pid(pid):        
    """ Check For the existence of a unix pid. """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True

def read_and_forward_pty_output(fd, child_pid, queue, queueResponse, printStdout):
    max_read_bytes = 1024 * 20
    continue_reading = True
    while continue_reading:
        time.sleep(0.01)
        if queue is not None and queue.qsize() > 0:
            key = queue.get(block=False)
            if key == "kill":
                parent = psutil.Process(child_pid)
                for child in parent.children(recursive=True):
                    child.kill()
                parent.kill()
                break
            else:
                try:
                    os.write(fd, key.encode())
                except OSError:
                    pass
        if fd:
            (data_ready, _, _) = select.select([fd], [], [], 0.5)
            if data_ready:
                try:
                    output = os.read(fd, max_read_bytes).decode(
                        errors="ignore"
                    )
                
                    output = output.replace("\r","")
                    if queue is not None:
                        if "Y/n" in output:
                            queue.put("Y\n", block=False)
                        elif "y/N" in output:
                            queue.put("N\n", block=False)
                    if printStdout:
                        print(output)
                    if queueResponse is not None:
                        queueResponse.put(output, block=False)
                except OSError:
                    break
    if fd:
        os.close(fd)
    try:
        if queue:
            while queue.qsize() > 0:
                queue.get(block=False)
            queue.close()
            queue.join_thread()
        if queueResponse:
            while queueResponse.qsize() > 0 and queueResponse.empty() is False:
                queueResponse.get(block=False)
            queueResponse.close()
            queueResponse.join_thread()
    except Exception as e:
        logger.error(f"Utils execute: read_and_forward_pty_output {e}")
    continue_reading = False
    return

def execute(command, timeout=None, queue=None, queueResponse=None, cwd=None, printStdout=False):
    """
    Execute a bash command and print output

    Args:
        command: A bash command
        timeout: a date in the futur when the command will be stopped if still running or None to not use this option, default as None.

    Returns:
        Return the return code of this command

    Raises:
        Raise a KeyboardInterrupt if the command was interrupted by a KeyboardInterrupt (Ctrl+c)
    """
    if isinstance(timeout, int):
        timeout = float(timeout)
    elif isinstance(timeout, float):
        timeout = timeout
    elif timeout is not None:
        if timeout.year < datetime.now().year+1:
            timeout = (timeout-datetime.now()).total_seconds()
        else:
            timeout = None
    
    (child_pid, fd) = pty.fork()
    if child_pid == 0:
        try:
            proc = subprocess.run(
                command, shell=True, cwd=cwd, timeout=timeout) #, preexec_fn=os.setsid
            sys.exit(0)
        except subprocess.TimeoutExpired:
            logger.debug("Timeout expired for command "+str(command))
            sys.exit(-1)
        except Exception as e:
            logger.debug("Exception occured for command "+str(command))
            sys.exit(1)
        finally:
            logger.debug("Exit finally for command "+str(command))
            sys.exit(0)
    else:
        
        p = multiprocessing.Process(target=read_and_forward_pty_output, args=[fd, child_pid, queue, queueResponse, printStdout])
        p.start()
        p.join()
    info = os.waitpid(child_pid, 0)
    try:
        if os.WIFEXITED(info[1]):
            returncode = os.WEXITSTATUS(info[1])
        else:
            returncode = -1
    except Exception as e:
        return -1
    return returncode




def execute_no_fork(command, timeout=None, printStdout=True, queue=None, queueResponse=None, cwd=None):
    """
    Execute a bash command and print output

    Args:
        command: A bash command
        timeout: a date in the futur when the command will be stopped if still running or None to not use this option, default as None.
        printStdout: A boolean indicating if the stdout should be printed. Default to True.

    Returns:
        Return the return code of this command

    Raises:
        Raise a KeyboardInterrupt if the command was interrupted by a KeyboardInterrupt (Ctrl+c)
    """
    
    try:
        time.sleep(1) #HACK Break if not there when launching fast custom tools on local host for unknown reason

        proc = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE, shell=True, cwd=cwd)
        proc._killed = False
        signal.signal(signal.SIGINT, lambda _signum, _frame: handleProcKill(proc))
        signal.signal(signal.SIGTERM, lambda _signum, _frame: handleProcKill(proc))
        try:
            timer = None
            if timeout is not None:
                if isinstance(timeout, int):
                    timeout = float(timeout)
                if isinstance(timeout, float):
                    timer = Timer(timeout, proc.kill)
                    timer.start()
                    logger.debug("Utils execute: timer start "+str(timeout))
                else:
                    if timeout.year < datetime.now().year+1:
                        timeout = (timeout-datetime.now()).total_seconds()
                        timer = Timer(timeout, proc.kill)
                        timer.start()
                        logger.debug("Utils execute: timer start "+str(timeout))
                    else:
                        timeout = None
            logger.debug(f"Utils execute: timeout:{timeout} command:{command}")
            output = b""
            os.set_blocking(proc.stdout.fileno(), False)
            os.set_blocking(proc.stdin.fileno(), False)
            while proc.poll() is None and queue is not None and queueResponse is not None:
                if queue is not None and queue.qsize() > 0:
                    print("queue not empty")
                    key = queue.get(block=False)
                    if not key:
                        continue
                    proc.stdin.write(key.encode())
                    data = proc.stdout.read()
                    queueResponse.put(data)
                    time.sleep(3)
            stdout, stderr = proc.communicate(None, timeout)
            if queueResponse is not None:
                queueResponse.put(stdout)
            if proc._killed:
                logger.debug(f"Utils execute: command killed command:{command}")
                if timer is not None:
                    timer.cancel()
                return -1, ""
            if printStdout:
                stdout = stdout.decode('utf-8')
                stderr = stderr.decode('utf-8')
                if str(stdout) != "":
                    print(str(stdout))
                if str(stderr) != "":
                    print(str(stderr))
        except Exception as e:
            import traceback
            logger.debug(f"Utils execute: command ended with exception {repr(e)}")
            traceback.print_exc(file=sys.stdout)
            print(repr(e))
            proc.kill()
            return -1, ""
        finally:
            if timeout is not None:
                if isinstance(timeout, float):
                    timer.cancel()
                else:
                    if timeout.year < datetime.now().year+1:
                        timer.cancel()
        return proc.returncode, stdout
    except KeyboardInterrupt as e:
        raise e


def get_screen_where_widget(widget):
    """
    Get the screen size of the current terminal

    Returns:
        Return the screen size as a tuple (width, height)
    """
    monitors = screeninfo.get_monitors()
    monitor = None
    if widget is not None:
        x = widget.winfo_rootx()
        y = widget.winfo_rooty()
        for m in reversed(monitors):
            if m.x <= x <= m.width + m.x and m.y <= y <= m.height + m.y:
                monitor = m
    if monitor is None:
        monitor = monitors[0]
    return monitor

def performLookUp(domain):
    """
    Uses the socket module to get an ip from a domain.

    Args:
        domain: the domain to look for in dns

    Returns:
        Return the ip found from dns records, None if failed.
    """
    try:
        return socket.gethostbyname(domain)
    except socket.gaierror:
        return None


def loadCfg(cfgfile):
    """
    Load a json config file.
    Args:
        cfgfile: the path to a json config file
    Raises:
        FileNotFoundError if the given file does not exist
    Returns:
        Return the json converted values of the config file.
    """
    cf_infos = dict()
    try:
        with open(cfgfile, mode="r") as f:
            cf_infos = json.loads(f.read())
    except FileNotFoundError as e:
        raise e
    except json.JSONDecodeError as e:
        raise e
    return cf_infos

def getConfigFolder():
    home = expanduser("~")
    config = os.path.join(home,".config/pollenisator-gui/")
    return config

def loadClientConfig():
    """Return data converted from json inside config/client.cfg
    Returns:
        Json converted data inside config/client.cfg
    """
    config = os.path.join(getConfigFolder(), "client.cfg")
    try:
        res = loadCfg(config)
        return res

    except:
        return {"host":"127.0.0.1", "port":"5000", "https":"False"}

def saveClientConfig(configDict):
    """Saves data in configDict to config/client.cfg as json
    Args:
        configDict: data to be stored in config/client.cfg
    """
    config_folder = getConfigFolder()
    try:
        os.makedirs(config_folder)
    except:
        pass
    configFile = os.path.join(config_folder, "client.cfg")
    with open(configFile, mode="w") as f:
        f.write(json.dumps(configDict))
  

def getValidMarkIconPath():
    """Returns:
         a validation mark icon path
    """
    return getIcon("done_tool.png")


def getBadMarkIconPath():
    """Returns:
         a bad mark icon path
    """
    return getIcon("cross.png")


def getWaitingMarkIconPath():
    """Returns:
         a waiting icon path
    """
    return getIcon("waiting.png")

def getIcon(name):
    """Returns : the path to an specified icon name
    """
    settings = cacheSettings()
    if settings.is_dark_mode():
        return os.path.join(getIconDir(), "dark", os.path.basename(name))
    else:
        return os.path.join(getIconDir(), "light", os.path.basename(name))
        
def getHelpIconPath():
    """Returns:
         a help icon path
    """
    return getIcon("help.png")

def loadIcon(name, **kwargs):
    """Returns:
         a help icon path
    """
    path = getIcon(name)
    image = Image.open(path)
    #img = image.resize((16, 16))
    return ImageTk.PhotoImage(image)

def loadIcons(names):
    load_icons = {}
    for name in names:
        load_icons[name] = loadIcon(name)
    return load_icons

def getIconDir():
    """Returns:
        the icon directory path
    """
    p = os.path.join(os.path.dirname(
        os.path.realpath(__file__)), "../../theme/")
    return p

def getLocalDir():
    """Returns:
        the pollenisator local folder
    """
    p = getMainDir()+"local/"
    return p

def getExportDir():
    """Returns:
        the pollenisator export folder
    """
    p = os.path.join(os.path.dirname(
        os.path.realpath(__file__)), "../../exports/")
    return p

def getMainDir():
    """Returns:
        the pollenisator main folder
    """
    p = os.path.join(os.path.dirname(
        os.path.realpath(__file__)), "../../")
    return p

def getColorTheme():
    return getMainDir()+"theme/pollenisator_theme.json"

def loadTheme():
    with open(getColorTheme(), "r") as f:
        return json.load(f)


def drop_file_event_parser(event):
    """Parse event Callback of python-tkdnd on file drop event
    event.data is built weirdly:
        - Each file dropped in will be space-separated in a long string.
        - If a directory/file contains a space, the whole filename will be wrapped by curly brackets.
        - If a filename contains a curly brackets it will be escaped by backslashes and
        - If a filename contains an opening and a closing curly brackets, they might not be escaped
        - If a filename contains a space and a curly brackets, the filename is is not wrapped by curly brackets.
            but the space and curly brackets will be escaped by backslashes
        Returns:
            list of valid filename for python
        Exceptions:
            raise FileNotFoundError if a filename is not valid
    """
    parts = event.data.split(" ")
    data = []
    cumul = ""
    expect_closing_bracket = False
    for part in parts:
        if part.startswith("{") and not expect_closing_bracket:
            cumul += part[1:]+" "
            expect_closing_bracket = True
        elif part.endswith("\\"):
            cumul += part+" "
        elif part.endswith("}") and not part.endswith("\\}") and expect_closing_bracket:
            cumul += part[:-1]
            data.append(cumul)
            cumul = ""
            expect_closing_bracket = False
        else:
            if expect_closing_bracket:
                cumul += part+" "
            else:
                cumul += part
                data.append(cumul)
                cumul = ""
    # check existance 
    sanitized_path = []
    for d in data:
        # remove espacing as python does not expect spaces and brackets to be espaced
        d = d.replace("\\}", "}").replace("\\{", "{").replace("\\ "," ")
        if not os.path.exists(d):
            raise FileNotFoundError(d)
        sanitized_path.append(d)
    return sanitized_path

def openPathForUser(path, folder_only=False):
    path_to_open = os.path.dirname(path) if folder_only else path
    cmd = ""
    if which("xdg-open"):
        cmd = "xdg-open "+path_to_open
    elif which("explorer"):
        cmd = "explorer "+path_to_open
    elif which("open"):
        cmd = "open "+path_to_open
    if cmd != "":
        subprocess.Popen(shlex.split(cmd))
    else: # windows
        try: 
            os.startfile(path_to_open)
        except Exception:
            return False
    return True

# def executeInExternalTerm(command, with_bash=True, default_target=None):
#     from pollenisatorgui.core.components.settings import Settings
#     settings = Settings()
#     favorite = settings.getFavoriteTerm()
#     if favorite is None:
#         tk.messagebox.showerror(
#             "Terminal settings invalid", "None of the terminals given in the settings are installed on this computer.")
#         return False
#     if which(favorite) is not None:
#         env = {}
#         if default_target is not None:
#             env["POLLENISATOR_DEFAULT_TARGET"] = default_target
#         env = {**os.environ, **env}
#         terms = settings.getTerms()
#         terms_dict = {}
#         for term in terms:
#             terms_dict[term.split(" ")[0]] = term
#         command_term = terms_dict.get(favorite, None)
#         if command_term is not None:
#             if not command_term.endswith(" "):
#                 command_term += " "
#             command_term += command
#             print(command_term)
#             subprocess.Popen(command_term, shell=True, env=env, cwd=getExportDir())
#         else:
#             tk.messagebox.showerror(
#                 "Terminal settings invalid", "Check your terminal settings")
#     else:
#         tk.messagebox.showerror(
#             "Terminal settings invalid", f"{favorite} terminal is not available on this computer. Choose a different one in the settings module.")
#     return True

def which_expand_aliases(whats):
    whats = list(whats)
    res = expand_alias(shlex.join(whats))
    if res is None:
        return res
    aliases = {}
    for i, result in enumerate(res.split("\n")):
        if "not found" in result:
            aliases[whats[i]] = None
        elif "aliased to " in result:
            aliases[whats[i]] = result.split("aliased to ")[1].strip()
        else:
            aliases[whats[i]] = result
    return aliases

def which_expand_alias(what):
    res = which(what)
    if res is not None:
        return res
    stdout = expand_alias(what)
    if stdout is None:
        return None
    if "not found" in stdout:
        return None
    elif "aliased to " in stdout:
        return stdout.split("aliased to ")[1].strip()
    else:
        return stdout.strip()

def expand_alias(what):
    settings = cacheSettings()
    settings.reloadLocalSettings()
    is_there_zsh = os.environ.get("ZSH",None) is not None
    default_shell = "zsh" if is_there_zsh else "/bin/bash"
    terminal = settings.local_settings.get("terminal", os.environ.get("SHELL",default_shell))
    rc_file = settings.local_settings.get("rc_file", "")
    if rc_file == "":
        home = expanduser("~")
        rc_file = os.path.join(home,"."+os.path.basename(terminal)+"rc") # rc file is not loaded automatically
    proc = subprocess.run(f"source {rc_file} && which {what}", executable=terminal, shell=True, stdout=subprocess.PIPE)
    if proc.returncode == 0 or proc.returncode == 1:
        stdout = proc.stdout.decode("utf-8")
        return stdout.strip()
    return None

def is_json(myjson):
  try:
    json.loads(myjson)
  except ValueError as e:
    return False
  return True


"""View for multi checkinstance list object. Present a form to user when interacted with."""

from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.views.viewelement import ViewElement
from customtkinter import *
from PIL import ImageTk, Image
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.components.settings import Settings
import tkinter.ttk as ttk
import tkinter as tk


class CheckInstanceMultiView(ViewElement):
    """View for checlist multi object. Present an multi  form to user when interacted with."""


    def __init__(self, appliTw, appViewFrame, mainApp):
        super().__init__(appliTw, appViewFrame, mainApp, None)

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or perform actions on multiple different objects common properties like tags.
        """
        self.form.clear()
        top_panel = self.form.addFormPanel()
        top_panel.addFormButton("Mark as done", self.markDoneSelection)
        top_panel.addFormButton("Mark as not done", self.markNotDoneSelection)
        top_panel.addFormButton("Export", self.appliTw.exportSelection)
        self.delete_image = CTkImage(Image.open(utils.getIcon("delete.png")))
        top_panel.addFormButton("Delete", self.appliTw.deleteSelected, image=self.delete_image,
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        selected = self.appliTw.selection()
        dones = []
        not_dones = []
        runnings = []
        for iid in selected:
            view = self.appliTw.views.get(iid, {}).get("view")
            if view is None:
                continue
            if view.controller.getStatus() == "done":
                dones.append(iid)
            elif view.controller.getStatus() == "running":
                runnings.append(iid)
            else:
                not_dones.append(iid)

        if not_dones:
            top_panel.addFormButton("Queue Selection", self.queueSelection)
        self.showForm()

    def markDoneSelection(self, _event=None):
        selected = self.appliTw.selection()
        if selected:
            apiclient = APIClient.getInstance()
            apiclient.multiChangeStatus(selected, "done")
    
    def markNotDoneSelection(self, _event=None):
        selected = self.appliTw.selection()
        if selected:
            apiclient = APIClient.getInstance()
            apiclient.multiChangeStatus(selected, "todo")

    def queueSelection(self, _event = None):
        selected = self.appliTw.selection()
        if selected:
            apiclient = APIClient.getInstance()
            results = apiclient.queueCheckInstances(selected, force=True)
            if len(results.get("successes", [])) < 10:
                for iid in results.get("successes", []):
                    self.mainApp.subscribe_notification("tool_start", self.toolStartedEvent, pentest=apiclient.getCurrentPentest(), iid=iid)
                tk.messagebox.showinfo("Info", "Tasks queued successfully. They should appear in the terminal view below when started", parent=self.mainApp)
            else:
                tk.messagebox.showinfo("Info", "Tasks queued successfully. There is too many to get notified on start.", parent=self.mainApp)
            if results.get("failures", []):
                failure = results.get("failures", [])[0]
                tk.messagebox.showerror("Error", "Error while queueing task : "+str(failure.get("error", "unknown error")), parent=self.mainApp)
                  


# """User auth infos module to store and use user login informations """
# import tkinter as tk
# import tkinter.messagebox
# import tkinter.ttk as ttk
# from customtkinter import *
# from bson.objectid import ObjectId
# import pollenisatorgui.core.components.utils as utils
# from pollenisatorgui.core.components.apiclient import APIClient
# from pollenisatorgui.core.application.dialogs.ChildDialogProgress import ChildDialogProgress
# from pollenisatorgui.modules.module import Module
#from pollenisatorgui.core.application.pollenisatorentry import PopoEntry

# class AuthInfo(Module):
#     """
#     Shows information about ongoing pentest. 
#     """
#     iconName = "tab_auth.png"
#     tabName = "Auth Infos"
#     coll_name = "auth"
#     order_priority = Module.LOW_PRIORITY
#     pentest_types = ["web"]

#     def __init__(self, parent, settings):
#         """
#         Constructor
#         """
#         super().__init__()
#         self.dashboardFrame = None
#         self.parent = None
#         self.treevw = None
#         self.style = None
#         self.icons = {}
    
#     def open(self):
#         apiclient = APIClient.getInstance()
#         if apiclient.getCurrentPentest() is not None:
#             self.refreshUI()
#         return True

#     def refreshUI(self):
#         """
#         Reload data and display them
#         """
#         self.loadData()
#         self.displayData()

#     def loadData(self):
#         """
#         Fetch data from database
#         """
#         apiclient = APIClient.getInstance()
#         self.auth = apiclient.find("auth")

#     def displayData(self):
#         """
#         Display loaded data in treeviews
#         """
#         dialog = ChildDialogProgress(self.parent, "Loading infos ",
#                                      "Refreshing infos. Please wait for a few seconds.", 200, "determinate")
#         dialog.show(3)

#         # Reset Ip treeview
#         for children in self.treevw.get_children():
#             self.treevw.delete(children)
#         dialog.update(1)
#         for auth_d in self.auth:
#             keys = list(auth_d.keys())

#             self.treevw.insert(
#                 '', 'end', auth_d["_id"], text=auth_d["name"], values=(auth_d["value"], auth_d["type"],))
#         dialog.update(3)
#         # Reset Port treeview
#         dialog.destroy()

#     def initUI(self, parent, nbk, treevw, tkApp):
#         """
#         Initialize Dashboard widgets
#         Args:
#             parent: its parent widget
#         """
#         if self.parent is not None:  # Already initialized
#             return
#         self.parent = parent
#         self.tkApp = tkApp
#         self.treevwApp = treevw
#         self.moduleFrame = CTkFrame(parent)

#         self.rowHeight = 20
#         self.style = ttk.Style()
#         self.style.configure('AuthInfo.Treeview', rowheight=self.rowHeight)

#         frameTwUsers = CTkFrame(self.moduleFrame)
#         self.treevw = ttk.Treeview(
#             frameTwUsers, style='AuthInfo.Treeview', height=10)
#         self.treevw['columns'] = ('name', 'value', 'type')
#         self.treevw.heading("#0", text='Name', anchor=tk.W)
#         self.treevw.column("#0", anchor=tk.W, width=50)
#         self.treevw.heading('value', text='Value')
#         self.treevw.column('value', anchor=tk.W, width=50)
#         self.treevw.heading('type', text='Type')
#         self.treevw.column('type', anchor=tk.W, width=30)
#         self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
#         self.treevw.bind('<Delete>', self.onAuthDelete)
#         scbVSel = CTkScrollbar(frameTwUsers,
#                                 orientation=tk.VERTICAL,
#                                 command=self.treevw.yview)
#         self.treevw.configure(yscrollcommand=scbVSel.set)
#         scbVSel.grid(row=0, column=1, sticky=tk.NS)
#         #frameTwHosts.pack(side=tk.TOP, fill=tk.X, padx=100, pady=10)
#         frameTwUsers.columnconfigure(0, weight=1)
#         frameTwUsers.rowconfigure(0, weight=1)
#         frameTwUsers.grid(row=0, column=0, sticky=tk.NSEW, padx=5, pady=10)
#         frameActions = CTkFrame(self.moduleFrame)
#         self.text_add_users = CTkTextbox(
#             frameActions, wrap="word")
#         self.text_add_users.pack(side=tk.TOP, padx=10,pady=10,fill=tk.X)
#         self.typeCombo = CTkComboBox(frameActions, values=("Cookie", "Password", "Hash"), state="readonly")
#         self.typeCombo.set("Cookie")
#         self.typeCombo.pack(side=tk.TOP)
#         lblDesc = CTkLabel(frameActions, text="Format is Name:Value (UserName:Cookie or Login:password for exemple. Each entry on a new line)")
#         lblDesc.pack(side=tk.TOP)
#         addBtn = CTkButton(frameActions, text="Add auth info", command=self.insert_auths)
#         addBtn.pack(side=tk.TOP)
#         frameActions.grid(row=0, column=1, sticky=tk.NSEW, padx=5, pady=10)
#         self.moduleFrame.columnconfigure(0, weight=1)
#         self.moduleFrame.columnconfigure(1, weight=1)
#         self.moduleFrame.pack(padx=10, pady=10, fill=tk.BOTH, expand=True)
     
#     def onAuthDelete(self, event):
#         """Callback for a delete key press on a worker.
#         Force deletion of worker
#         Args:
#             event: Auto filled
#         """
#         apiclient = APIClient.getInstance()
#         selected = self.treevw.selection()
#         apiclient.bulkDelete({"auth":selected}) 


#     def insert_auths(self):
#         apiclient = APIClient.getInstance()
#         toInsert = []
#         for line in self.text_add_users.get('1.0', tk.END).split("\n"):
#             if line.strip() == "":
#                 continue
#             parts = line.strip().split(":")
#             if len(parts) < 2:
#                 tkinter.messagebox.showerror("Syntax error", f"Format expected is name:value (line : {line})")
#                 return
#             toInsert.append([parts[0],(":".join(parts[1:]))])
#         print(toInsert)
#         for t in toInsert:
#             apiclient.insert("auth", {"name": t[0], "value":t[1], "type":self.typeCombo.get().lower()})

#     def update_received(self, dataManager, notif, obj, old_obj):
#         apiclient = APIClient.getInstance()
#         if apiclient.getCurrentPentest() != notif["db"]:
#             return
#         iid =  notif["iid"]
#         if notif["action"] == "insert":	
#             res = apiclient.find("auth", {"_id": ObjectId(iid)}, False)	
#             if res is None:
#                 return
#             self.treevw.insert(
#                 '', 'end', res["_id"], text=res["name"], values=(res["value"], res["type"],))
#         elif notif["action"] == "delete":
#             try:
#                 self.treevw.delete(str(iid))
#             except tk.TclError:
#                 pass

#     def _initContextualsMenus(self, parentFrame):
#         """
#         Create a contextual menu
#         """
#         self.contextualMenu = utils.craftMenuWithStyle(parentFrame)
#         self.contextualMenu.add_command(
#             label="Add Credentials", command=self.addCredentials)
#         return self.contextualMenu

#     def addCredentials(self, _event=None):
#         dialog = ChildDialogAuthInfo(self.tkApp)
#         self.tkApp.wait_window(dialog.app)
#         if isinstance(dialog.rvalue , dict):
#             apiclient = APIClient.getInstance()
#             name = dialog.rvalue["key"].strip()
#             value = dialog.rvalue["value"].strip()
#             auth_info = {"name": name, "value":value, "type":dialog.rvalue["type"].lower()}
#             res, iid = apiclient.insert("auth", auth_info)
#             for selected in self.treevwApp.selection():
#                 view_o = self.treevwApp.getViewFromId(selected)
#                 if view_o is not None:
#                     res = apiclient.linkAuth(str(iid), view_o.controller.getDbId())

# class ChildDialogAuthInfo:
#     """
#     Open a child dialog of a tkinter application to ask a user a pentest name.
#     """
#     def __init__(self, parent, displayMsg="Add a key value auth info (login:password) (cookie_name:value)"):
#         """
#         Open a child dialog of a tkinter application to ask a combobox option.

#         Args:
#             parent: the tkinter parent view to use for this window construction.
#             displayMsg: The message that will explain to the user what he is choosing.
#             default: Choose a default selected option (one of the string in options). default is None
#         """
#         self.app = CTkToplevel(parent, fg_color=utils.getBackgroundColor())
#         self.app.title("Auth information")
#         self.app.resizable(False, False)
#         appFrame = CTkFrame(self.app)
#         self.rvalue = None
#         self.parent = parent
#         lbl = CTkLabel(appFrame, text=displayMsg)
#         lbl.pack(pady=5)
#         valFrame = CTkFrame(self.app)
#         self.box = CTkComboBox(
#             valFrame, values=tuple(["Password","Cookie"]), state="readonly")
#         self.box.bind('<<ComboboxSelected>>', self.box_modified)  
#         self.box.set("Password")
#         self.box.grid(column=0, row=0)
#         self.lbl_key = CTkLabel(valFrame, text="Login:")
#         self.lbl_key.grid(row=0, column=1)
#         self.entry_key = PopoEntry(valFrame)
#         self.entry_key.grid(row=0, column=2)
#         self.lbl_value = CTkLabel(valFrame, text="Password:")
#         self.lbl_value.grid(row=0, column=3)
#         self.entry_value = PopoEntry(valFrame)
#         self.entry_value.grid(row=0, column=4)
#         valFrame.pack(ipadx=10, ipady=5, pady=10, padx=10)
#         self.ok_button = CTkButton(appFrame, text="OK", command=self.onOk)
#         self.ok_button.bind('<Return>', self.onOk)
#         self.ok_button.pack(padx=10, pady=5, side="right")
#         self.cancel_button = CTkButton(appFrame, text="Cancel", command=self.onError, 
#                                fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
#                                border_width=1, border_color="firebrick1", hover_color="tomato")
#         self.cancel_button.pack(padx=10, pady=5, side="right")
#         appFrame.pack(ipadx=10, ipady=5)
#         try:
#             self.app.wait_visibility()
#             self.app.transient(parent)
#             self.app.grab_set()
#             self.app.focus_force()
#             self.app.lift()
#         except tk.TclError:
#             pass
#         self.box.after(50, self.openCombo)

#     def openCombo(self):
#         self.box.event_generate('<Button-1>')

#     def box_modified(self, event=""):
#         if self.box.get() == "Password":
#             self.lbl_key.configure(text="Login:")
#             self.lbl_value.configure(text="Password:")
#         elif self.box.get() == "Cookie":
#             self.lbl_key.configure(text="Name:")
#             self.lbl_value.configure(text="Value:")

#     def onOk(self, event=""):
#         """
#         Called when the user clicked the validation button. Set the rvalue attributes to the value selected and close the window.
#         """
#         # send the data to the parent
#         self.rvalue = {"type":self.box.get(), "key":self.entry_key.get(), "value":self.entry_value.get()}
#         self.app.destroy()

#     def onError(self):
#         """
#         Close the dialog and set rvalue to None
#         """
#         self.rvalue = None
#         self.app.destroy()



from pollenisatorgui.core.plugins.plugin import Plugin
from pollenisatorgui.core.models.port import Port
import webbrowser

class Nikto(Plugin):
    def __init__(self):
        self.toolmodel = None
    def getActions(self, toolmodel):
        """
        Summary: Add buttons to the tool view.
        Args:
            * toolmodel : the tool model opened in the pollenisator client.
        Return:
            A dictionary with buttons text as key and function callback as value.
        """
        self.toolmodel = toolmodel
        return {"Open in browser": self.openInBrowser}

    def openInBrowser(self, _event=None):
        """Callback of action  Open 200 in browser
        Open scanned host port in browser as tabs.
        Args:
            _event: not used but mandatory
        """
        port_m = Port.fetchObject(
            {"ip": self.toolmodel.ip, "port": self.toolmodel.port, "proto": self.toolmodel.proto})
        if port_m is None:
            return
        ssl = port_m.infos.get("SSL", None) == "True" or ("https" in port_m.service or "ssl" in port_m.service)
        url = "https://" if ssl else "http://"
        url += port_m.ip+":"+str(port_m.port)+"/"
        webbrowser.open_new_tab(url)


"""Describe tkinter switch with default common args"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.forms.form import Form


class FormSwitch(Form):
    """
    Form field representing a switch.
    Default setted values: 
        if pack : padx = pady = 5, side = right
        if grid: row = column = 0 sticky = "west"
    Kwargs:
        binds:  a dictionnary of tkinter binding with shortcut as key and callback as value
    """
    def __init__(self, name, text, default, **kwargs):
        """
        Constructor for a form switch
        
        Args:
            name: the switch name (id).
            text: the text on the switch
            default: boolean indicating if the switch should be checked by default.
            kwargs: same keyword args as you would give to CTkSwitch
        """
        super().__init__(name)
        self.text = text
        self.default = bool(default)
        self.kwargs = kwargs
        self.chk = None

    def constructView(self, parent):
        """
        Create the switch view inside the parent view given

        Args:
            parent: parent form panel.
        """
        self.val = tk.IntVar()
        if self.default:
            self.val.set(1)
        else:
            self.val.set(0)
        self.chk = CTkSwitch(
            parent.panel, text=self.text, variable=self.val)
        binds = self.getKw("binds", {})
        for bind in binds:
            self.chk.bind(bind, binds[bind])
        command = self.getKw("command", None)
        if command is not None:
            self.chk.configure(command=command)
        
        if parent.gridLayout:
            self.chk.grid(row=self.getKw("row", 0), column=self.getKw("column", 0), sticky=self.getKw("sticky", tk.W), **self.kwargs)
        else:
            self.chk.pack(side=self.getKw("side", "right"), padx=self.getKw("padx", 10), pady=self.getKw("pady", 5), **self.kwargs)

    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return True if the switch was checked, False otherwise.
        """
        return self.val.get() == 1

    def checkForm(self):
        """
        Check if this form is correctly filled. A switch cannot be malformed.

        Returns:
            {
                "correct": True if the form is correctly filled, False otherwise.
                "msg": A message indicating what is not correctly filled.
            }
        """
        return True, ""

    def setFocus(self):
        """Set the focus to the ttk checkbutton."""
        self.chk.focus_set()

    


"""Controller for remark object. Mostly handles conversion between mongo data and python objects"""

import os
from pollenisatorgui.core.controllers.controllerelement import ControllerElement


class RemarkController(ControllerElement):
    """Inherits ControllerElement
    Controller for remark object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Remark represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by RemarkView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated Remark document.
        """
        self.model.title = values.get("Title", self.model.title)
        self.model.type = values.get("Type", self.model.type)
        self.model.description = values.get("Description", self.model.description)
        # Updating
        self.model.update()

    def doInsert(self, values):
        """
        Insert the Remark represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by RemarkView containing all form fields values needed.

        Returns:
            {
                '_id': The mongo ObjectId _id of the inserted command document.
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        title = values["Title"]
        typeof = values["Type"]
        description = values["Description"]
        self.model.initialize(typeof, title, description)
        ret, _ = self.model.addInDb()

        return ret, 0  # 0 erros

    
    def getType(self):
        """Returns a string describing the type of object
        Returns:
            "remark" """
        return "remark"


#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@author: Fabien Barré for AlgoSecure
# Date: 12/10/2022
# Major version released: 09/2019
# @version: 2.2
"""

import time
import tkinter as tk
import os
import sys
import git
import signal
import requests
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.application.appli import Appli
import pollenisatorgui.core.components.utils as utils
import threading
from getpass import getpass
from pollenisatorgui.core.components.logger_config import logger
import customtkinter
import subprocess

event_obj = threading.Event()
customtkinter.set_default_color_theme(utils.getColorTheme())

class GracefulKiller:
    """
    Signal handler to shut down properly.

    Attributes:
        kill_now: a boolean that can checked to know that it's time to stop.
    """
    kill_now = False

    def __init__(self, app):
        """
        Constructor. Hook the signals SIGINT and SIGTERM to method exitGracefully

        Args:
            app: The appli object to stop
        """
        signal.signal(signal.SIGINT, self.exitGracefully)
        signal.signal(signal.SIGTERM, self.exitGracefully)
        # signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        self.app = app

    def exitGracefully(self, _signum, _frame):
        """
        Set the kill_now class attributes to True. Call the onClosing function of the application given at init.

        Args:
            signum: not used
            frame: not used
        """
        #import traceback
        #traceback.print_stack(_frame)
        self.app.onClosing()
        event_obj.set()
        self.kill_now = True


#######################################
############## MAIN ###################
#######################################

def get_local_git_version():
    try:
        repo = git.Repo(search_parent_directories=True)
        latest_tag = repo.tags[-1]
        latest_version = latest_tag.name
        return latest_version
    except Exception as e:
        print("Error fetching latest version from Git:", e)
        return None


def get_latest_github_version():
    try:
        # Make a GET request to GitHub's releases endpoint
        url = f"https://api.github.com/repos/fbarre96/PollenisatorGUI/releases/latest"
        response = requests.get(url)
        if response.status_code == 200:
            latest_version = response.json()["tag_name"]
            return latest_version
        else:
            print("Error fetching latest version from GitHub:", response.text)
            return None
    except Exception as e:
        print("Error fetching latest version from GitHub:", e)
        return None
    
def checkForUpdates():
    latest = get_latest_github_version()
    local = get_local_git_version()
    if local is None or latest is None:
        return False
    if local != latest:
        return True
    return False

def update():
    try:
        # Use pipx to install the latest version of the package
        subprocess.run(['pipx', 'upgrade', 'pollenisatorgui'])
        print("Package updated successfully!")
    except Exception as e:
        print("Error updating package:", e)

def consoleConnect(force=False, askPentest=True):    
    apiclient = APIClient.getInstance()
    abandon = False
    if force:
        apiclient.disconnect()
    
    while (not apiclient.tryConnection(force=force) or not apiclient.isConnected()) and not abandon:
        success = promptForConnection() is None
    if apiclient.isConnected() and askPentest:
        promptForPentest()
    return not abandon
        
def promptForConnection():
   
    clientCfg = utils.loadClientConfig()
    host = clientCfg.get("host", "")
    port = clientCfg.get("port", "5000")
    https = str(clientCfg.get("https", "True")).title() == "True"

    host_result = input(f'Server hostname (defaut {host}):')
    if host_result.strip() == "":
        host_result = host
    port_result = input(f'Server Port (default  {port}):')
    if port_result.strip() == "":
        port_result = port
    https_result = input(f'Use HTTPS (y/N)?').strip() == 'y'
    
    apiclient = APIClient.getInstance()
    apiclient.reinitConnection()
    res = apiclient.tryConnection({"host":host_result, "port":port_result, "https":https_result})
    if res:
        username_result = input(f'Please input username:')
        password_result = getpass(prompt=f'Please input password:')
        #  pylint: disable=len-as-condition
        loginRes = apiclient.login(username_result, password_result)
        return loginRes
    else:
        print("Failed to contact this server.")
    return False

def promptForPentest():
    apiclient = APIClient.getInstance()
    pentests = apiclient.getPentestList()
    if pentests is None:
        pentests = []
    else:
        pentests.sort(key=lambda x: x["creation_date"], reverse=True)
    i = 0
    while i<=0:
        for i_p, pentest in enumerate(pentests):
            print(f"{i_p+1} : {pentest.get('nom', '')}")
        try:
            i = int(input("Selection (1-"+str(len(pentests))+"): "))
            if i <= 0 or i > len(pentests):
                print("Invalid selection, input must be between 1 and "+str(len(pentests)))
                i = 0
        except ValueError:
            i = 0
    apiclient.setCurrentPentest(pentests[i-1]["uuid"])

def parseDefaultTarget(stringToParse):
    if "|" in stringToParse:
        parts = stringToParse.split("|")
        ret = {}
        ret["check_iid"] = parts[0]
        ret["tool_iid"] = parts[1]
        ret["lvl"] = "import"
        return ret

    else:
        return {"check_iid":str(stringToParse), "lvl":"import"}


def pollup():
    """Send a file to pollenisator backend for analysis
    """
    if len(sys.argv) == 2:
        filename = sys.argv[1]
        plugin = "auto-detect"
    elif len(sys.argv) == 3:
        filename = sys.argv[1]
        plugin = sys.argv[2]
    else:
        print("Usage : pollup <filename> [plugin or auto-detect]")
        sys.exit(1)
    uploadFile(filename, plugin)

def uploadFile(filename, plugin='auto-detect'):
    apiclient = APIClient.getInstance()
    apiclient.tryConnection()
    res = apiclient.tryAuth()
    if not res:
        consoleConnect()
    print(f"INFO : Uploading results {filename}")
    msg = apiclient.importExistingResultFile(filename, plugin, parseDefaultTarget(os.environ.get("POLLENISATOR_DEFAULT_TARGET", "")), "")
    print(msg)

def pollwatch():
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler
    class Handler(FileSystemEventHandler):
        @staticmethod
        def on_any_event(event):
            if event.is_directory:
                return None
            elif event.event_type == 'created':
                # Event is created, you can process it now
                uploadFile(event.src_path)
    apiclient = APIClient.getInstance()
    apiclient.tryConnection()
    res = apiclient.tryAuth()
    if not res:
        consoleConnect()
    observer = Observer()
    observer.schedule(Handler(), '.', recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(5)
    finally:
        observer.stop()
        observer.join()

    

def main():
    """Main function. Start pollenisator application
    """
    
    print("""
.__    ..              ,       
[__) _ || _ ._ * __ _.-+- _ ._.
|   (_)||(/,[ )|_) (_] | (_)[  
                               
""")
    event_obj.clear() 
    gc = None
    update_available = checkForUpdates()
    if update_available:
        ask_update = input("An update is available, do you want to update now (y/N)?").strip().lower() == "y"
        if ask_update:
            update()
    app = Appli()
    try:
        gc = GracefulKiller(app)
        if not app.quitting:
            app.mainloop()
        print("Exiting tkinter main loop")
    except tk.TclError:
        pass
    try:
        try:
            app.destroy()
            print("Destroying app window")
        except tk.TclError:
            pass
        if gc is not None:
            gc.kill_now = True
        event_obj.set()
        app.onClosing()
    except Exception as e:
        print("Error while closing app")
        print(e)
if __name__ == '__main__':
    main()



"""View for wavr object. Handle node in treeview and present forms to user when interacted with."""
import tkinter as tk
from pollenisatorgui.core.controllers.checkinstancecontroller import CheckInstanceController
from pollenisatorgui.core.models.checkinstance import CheckInstance
from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView

from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.models.command import Command
from pollenisatorgui.core.models.interval import Interval
from pollenisatorgui.core.views.intervalview import IntervalView
from pollenisatorgui.core.models.scope import Scope
from pollenisatorgui.core.views.scopeview import ScopeView
from pollenisatorgui.core.views.toolview import ToolView
from pollenisatorgui.core.controllers.intervalcontroller import IntervalController
from pollenisatorgui.core.controllers.scopecontroller import ScopeController
from pollenisatorgui.core.controllers.toolcontroller import ToolController
from pollenisatorgui.core.components.apiclient import APIClient

from bson import ObjectId

class WaveView(ViewElement):
    """View for wavr object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory"""

    icon = 'wave.png'

    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Wave
        """
        self.form.clear()
        modelData = self.controller.getData()
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Scope options", modelData["wave"])
        # self.form.addFormHelper(
        #     "If you select a previously unselected command,\n it will be added to every object of its level.\nIf you unselect a previously selected command,\n it will remove only tools that are not already done.")
        # commands = Command.getList(None, APIClient.getInstance().getCurrentPentest())
        # commands_names = []
        # defaults = []
        # comms_values = []
        # for c in commands:
        #     commands_names.append(str(c))
        #     comms_values.append(c.getId())
        #     if str(c.getId()) in modelData["wave_commands"]:
        #         defaults.append(str(c))
        # self.form.addFormChecklist(
        #     "Commands", commands_names, defaults, values=comms_values)
        self.completeModifyWindow()

    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new Wave
        """
        top_panel = self.form.addFormPanel(grid=True)
        top_panel.addFormLabel("Scope options")
        top_panel.addFormStr("Wave", r".+", "", column=1)
        # self.form.addFormHelper("Only selected commands will be launchable.")
        # commands = Command.getList(None, APIClient.getInstance().getCurrentPentest())
        # commands_names = []
        # comms_values = []
        # for c in commands:
        #     commands_names.append(str(c))
        #     comms_values.append(c.getId())
        # self.form.addFormChecklist("Commands", commands_names, [], values=comms_values)
        self.completeInsertWindow()

    def getAdditionalContextualCommands(self):
        return {"Insert wave": self.openInsertWindow}

    def addChildrenBaseNodes(self, newNode):
        """
        Add to the given node from a treeview the mandatory childrens.
        For a wave it is the intervals parent node and the copes parent node.

        Args:
            newNode: the newly created node we want to add children to.
        Returns:
            * the created Intervals parent node
            * the created Scope parent node
        """
        d = self.appliTw.insert(newNode, "end", IntervalView.DbToTreeviewListId(
            self.controller.getDbId()), text="Intervals", image=IntervalView.getClassIcon())
        s = self.appliTw.insert(newNode, "end", ScopeView.DbToTreeviewListId(
            self.controller.getDbId()), text="Scopes", image=ScopeView.getClassIcon())
        return d, s

    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
            addChildren: If False: skip interval, tools and scope insert. Useful when displaying search results.
        """
        addChildren = kwargs.get("addChildren", True)
        self.appliTw.views[str(self.controller.getDbId())] = {
            "view": self, 'parent': ''}
        if parentNode is None:
            parentNode = self.getParentNode()
        try:
            wave_node = self.appliTw.insert(parentNode, "end", str(self.controller.getDbId()), text=str(
                self.controller.getModelRepr()), tags=self.controller.getTags(), image=self.getClassIcon())
        except tk.TclError:
            wave_node = str(self.controller.getDbId())
        if addChildren:
            dates_node, scopes_node = self.addChildrenBaseNodes(wave_node)
            intervals = self.controller.getIntervals()
            for interval in intervals:
                interval_vw = IntervalView(
                    self.appliTw, self.appliViewFrame, self.mainApp, IntervalController(Interval(interval)))
                interval_vw.addInTreeview(dates_node)
            checks = self.controller.getChecks()
            for check in checks:
                check_o = CheckInstanceController(check)
                check_vw = CheckInstanceView(self.appliTw, self.appliViewFrame, self.mainApp, check_o)
                check_vw.addInTreeview(str(self.controller.getDbId()))
            scopes = self.controller.getScopes()
            for scope in scopes:
                scope_o = ScopeController(Scope(scope))
                scope_vw = ScopeView(
                    self.appliTw, self.appliViewFrame, self.mainApp, scope_o)
                scope_vw.addInTreeview(scopes_node)
        if self.mainApp.settings.is_checklist_view():
            self.hide("checklist_view")
        if "hidden" in self.controller.getTags():
            self.hide("tags")

    def getParentNode(self):
        """
        Return the id of the parent node in treeview.

        Returns:
            return the saved waves_node inside the Appli class.
        """
        return self.appliTw.waves_node




"""StatusBar class. Show tagged elements numbers to user.
"""
import tkinter as tk
from tkinter import ttk
from customtkinter import *
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.components.tag import TagInfos

class StatusBar(CTkFrame):
    """StatusBar class. Show tagged numbers to user.
    Inherits CTkFrame
    """

    def tagClicked(self, name):
        """A lambda to callback with the tag name clicked
        Args:
            name: the tag name clicked
        """
        return lambda _event: self.statusbarClickedCallback(name)

    def __init__(self, master, statusbarClickedCallback):
        """
        Constructor of the status bar
        Args:
            master: parent tkinter window
            statusbarController: a controller to handle clicks on status bar.
                                It has to delcare a statusbarClicked function taking 1 arg : a tag name
        """
        # Cannot be imported at module level as tkinter will not have loaded.
        super().__init__(master, fg_color="transparent")
        # label = CTkLabel(self, text="Tagged")
        # label.pack(side="left")
        self.registeredTags = []
        self.statusbarClickedCallback = statusbarClickedCallback
        self.tagsCount = {}
        self.labelsTags = {}
        self.font = CTkFont("roboto", 11)
        DataManager.getInstance().attach(self)

    def refreshUI(self):
        for widget in self.winfo_children():
            widget.destroy()
        #self.pack_forget()
        lbl = CTkLabel(self, text="Filter by tags")
        lbl.pack(side="left", padx=1)

        self.registeredTags = Settings.getTags(ignoreCache=True)
        column = 1
        keys = list(self.registeredTags.keys())
        self.listOfLambdas = [self.tagClicked(keys[i]) for i in range(len(keys))]
        datamanager = DataManager.getInstance()
        tags = datamanager.get("tags", "*")
        for tag_obj in tags:
            for tag_info in tag_obj.tags:
                tag = TagInfos(tag_info)
                tag_name = tag.name
                self.tagsCount[str(tag_name)] = self.tagsCount.get(str(tag_name), 0) + 1
        for registeredTag, tag_info in self.registeredTags.items():
            self.tagsCount[registeredTag] = self.tagsCount.get(registeredTag, 0)
            color = tag_info["color"]
            try:
                if color == "transparent":
                    color = "white"
                self.labelsTags[registeredTag] = CTkLabel(self,  text=registeredTag+" : "+str(self.tagsCount[registeredTag]), fg_color=color, text_color="black",height=10, width=0, corner_radius=20,font=self.font)
            except tk.TclError:
                #color does not exist
                color = "gray97"
                self.labelsTags[registeredTag] = CTkLabel(self,text=registeredTag+" : "+str(self.tagsCount[registeredTag]),  fg_color=color, text_color="black",height=10, width=0,  corner_radius=20,font=self.font)
            self.labelsTags[registeredTag].pack(side="left", padx=1)
            self.labelsTags[registeredTag].bind('<Button-1>', self.listOfLambdas[column-1])
            column += 1

    def notify(self, addedTags, removedTags=[]):
        """
        Notify is called when tags are added or removed
        Args:
            addedTags: a list of tag names added
            removedTags: a list of tag names removed, default to []
        """
        updated = False
        if not "hidden" in addedTags:
            for tag in addedTags:
                tag = TagInfos(tag)
                if tag.name in self.tagsCount:
                    self.tagsCount[tag.name] += 1
                    updated = True
        for tag in removedTags:
            tag = TagInfos(tag)
            if tag.name in self.tagsCount:
                self.tagsCount[tag.name] -= 1
                updated = True
        if updated:
            self._update()
    

    def reset(self):
        """
        Rest all displayed tags count to 0
        """
        for registeredTag in self.registeredTags:
            self.tagsCount[registeredTag] = 0
        self._update()

    def _update(self):
        """
        Update all tags label to tags count
        """
        try:
            for tag, label in self.labelsTags.items():
                label.configure(text=tag+" : "+str(self.tagsCount.get(tag, 0)))
                label.update_idletasks()
        except Exception:
            pass

    def update_received(self, dataManager, notification, obj, old_obj):
        """
        Update the status bar when a tag is added or removed
        Args:
            dataManager: the DataManager instance
            notification: the notification type
            obj: the object that has been added or removed
        """
        if notification["collection"] == "settings":
            self.after(2000, self.refreshUI)
        if notification["collection"] == "tags":
            if old_obj is not None and obj is not None:
                if old_obj.tags != obj.tags:
                    self.notify(obj.tags, old_obj.tags)
    

    def refreshTags(self, tags):
        self.registeredTags = tags
        self.refreshUI()




from pollenisatorgui.core.models.port import Port
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.core.components.apiclient import APIClient


def main(apiclient):
	APIClient.setInstance(apiclient)
	apiclient.registerTag(apiclient.getCurrentPentest(), "unscanned", "yellow")
	ports = Port.fetchObjects({})
	n = 0
	for port in ports:
		port_key = port.getDbKey()
		res = Tool.fetchObject(port_key)
		if res is None:
			port.setTags(["unscanned"])
			n += 1

	return True, f"{n} ports found and marked as unscanned"
